/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.abstractfactory;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.TreeSet;

import dk.rode.thesis.bridge.SequenceValueArrayList;
import dk.rode.thesis.bridge.SequenceValueCollection;
import dk.rode.thesis.bridge.SequenceValueGenerator;
import dk.rode.thesis.bridge.SequenceValueHashSet;
import dk.rode.thesis.bridge.SequenceValueTreeSet;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.ParticipantUsage;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.util.Util;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>collection value factory</i> creates specific {@link SequenceValueCollection}
  * objects based on the supplied type of collection to the creation method. <p>  
  *
  * This factory is thread-safe. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values generated by created generators.
  */
@ParticipantUsage(value = "ConcreteProduct", type = SequenceValueGenerator.class)
@Participant("ConcreteFactory")
public class CollectionValueFactory<E extends Serializable> implements GeneratorFactory<E,Collection<E>> {

    /**
      * The type of values stored in the value collection. <p>
      * 
      * Never null.
      */
    private final Class<E> type;
    /**
      * The default collection to use in case the creation process is not
      * supplied an argument. Will be copied before each use. <p>
      * 
      * Never null.
      */
    private final ArrayList<E> defaultCollection;
    
    /**
      * Constructor. 
      * 
      * @param  type              The type of values generated by the created generators; cannot be null.
      * @param  defaultCollection The default collection to use in case the creation
      *                           process is not supplied an argument, but will be
      *                           copied before each use; cannot be null.
      * @throws NullPointerException If either argument is null.
      * @throws IllegalArgumentException if {@code defaultCollection} is empty.                           
      */
    public CollectionValueFactory(Class<E> type, Collection<E> defaultCollection) {
        if ((this.type = type) == null) {
            throw new NullPointerException("Type cannot be null");
        }
        // Create default as it will validate (and copy) "defaultCollection"... Have
        // to use a specific type as Collection is not serializable, only concrete
        // implementations of it...
        this.defaultCollection = new ArrayList<E>();
        // Minor hack to avoid logging...
        Log out = LogFactory.getLog(this);
        boolean active = out.isActive();
        out.setActive(false);
        this.createGenerator(defaultCollection).toCollection(this.defaultCollection);
        out.setActive(active);
    }
    
    /**
      * 
      */
    public Collection<E> getDefaultArgument() {
        return this.defaultCollection;
    }
    
    /**
      * 
      */
    public SequenceValueCollection<E,?> createGenerator() {
        return this.createGenerator(this.getDefaultArgument());
    }
    
    /**
      * 
      */
    public SequenceValueCollection<E,?> createGenerator(Collection<E> collection) {
        if (collection == null) {
            collection = Util.serializableCopy(this.defaultCollection);
        }
        SequenceValueCollection<E,?> values = null;
        if (collection instanceof ArrayList) {
            values = new SequenceValueArrayList<E>(this.type, (ArrayList<E>)collection);
            
        } else if (collection instanceof LinkedHashSet) {
            values = new SequenceValueHashSet<E>(this.type, (LinkedHashSet<E>)collection);

        } else if (collection instanceof TreeSet) {
            values = new SequenceValueTreeSet<E>(this.type, (TreeSet<E>)collection);
            
        } else if (collection instanceof HashSet) {
            values = new SequenceValueHashSet<E>(this.type, (HashSet<E>)collection);
            
        // Default: we use a list, since we do not want to remove possible 
        // duplicates, and the list will maintain the iteration order of 
        // "collection"...
        } else {
            values = new SequenceValueArrayList<E>(this.type, new ArrayList<E>(collection));
        }
        LogFactory.getLog(this).println("Created value generator: ", values);
        return values;
    }

    /**
      * Returns the string representation of this factory.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; default-collection: ").append(this.defaultCollection.getClass().getSimpleName()).append(this.defaultCollection).append(']').toString();
    }
    
}
