/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.abstractfactory;

import dk.rode.thesis.bridge.SequenceAbstraction;
import dk.rode.thesis.bridge.SequenceValueGenerator;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.ParticipantUsage;

/**
  * A <i>generator factory</i> can create sequence value 
  * {@link SequenceValueGenerator generators} used as implementations
  * for sequence {@link SequenceAbstraction abstractions}.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    SequenceFactory
  * @see    AbstractionFactory
  * 
  * @param  <E> The type of values generated by created generators.
  * @param  <P> The type of argument required in the creation process. Use
  *             {@code Void} if no arguments are required.
  */
@ParticipantUsage(value = "AbstractProduct", type = SequenceValueGenerator.class)
@Participant("AbstractFactory")
public interface GeneratorFactory<E,P> {

    /**
      * Returns the default argument to use for generator creation in case
      * no argument is supplied to the {@link #createGenerator(Object)}
      * method.
      *  
      * @return The default argument, if any; can be null, depending on the
      *         actual implementation.
      */
    public P getDefaultArgument();
    
    /**
      * Creates a new sequence value generator using the 
      * {@link #getDefaultArgument() default} argument. <p> 
      * 
      * @return A new created generator; never null.
      * @see    #createGenerator(Object)
      * @see    #getDefaultArgument()
      */
    public SequenceValueGenerator<E> createGenerator();
    
    /**
      * Creates a new sequence value generator, where {@code argument} is an 
      * argument used in the creation process of the generator. 
      * 
      * @param  argument A context supplied argument to use in the construction of the
      *                  generator, if any. If null, the {@link #getDefaultArgument()
      *                  default} argument is used (which may also be null, depending
      *                  on the actual implementation).
      * @return A new created generator; never null.
      * @see    #createGenerator()
      */
    public SequenceValueGenerator<E> createGenerator(P argument);
    
}
