/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.abstractfactory;

import static dk.rode.thesis.meta.util.Util.add;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.TreeSet;

import dk.rode.thesis.adapter.AdaptedPolicy;
import dk.rode.thesis.adapter.AdapterStrategy;
import dk.rode.thesis.adapter.SequenceAdapter;
import dk.rode.thesis.bridge.MemorizableSequenceAbstraction;
import dk.rode.thesis.bridge.SequenceAbstraction;
import dk.rode.thesis.bridge.SequenceValueRange;
import dk.rode.thesis.bridge.SynchronisedSequenceAbstraction;
import dk.rode.thesis.facade.FibonacciSequence;
import dk.rode.thesis.facade.MathFacade;
import dk.rode.thesis.factorymethod.Factory;
import dk.rode.thesis.factorymethod.TypedFactory;
import dk.rode.thesis.iterator.IterableSequence;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.LongSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.reflect.TypeLiteral;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.prototype.StrictCopyable;

/**
  * <b>Abstract Factory</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("Client")
public class Main implements Testable {
    
    /**
      * Executes the <b>Abstract Factory</b> tests.
      * 
      * @param args The arguments to supply to {@link #test(Log, Arguments)},
      *             if any; can be null.
      */
    public static void main(String[] args) {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }

    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * 
      */
    public boolean test(Log out, Arguments arguments) {
        out.heading("Abstract Factory tests...");
        
        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }

        boolean ok = true;

        //
        // Initialise logging
        //
        
        boolean log = arguments.getArgumentValue("log", true, Boolean.class);
        // Generator factories...
        LogFactory.getLog(CollectionValueFactory.class).setActive(log);
        LogFactory.getLog(RangeValueFactory.class).setActive(log);
        // Abstraction factories...
        LogFactory.getLog(StandardAbstractionFactory.class).setActive(log);
        LogFactory.getLog(MemorizableAbstractionFactory.class).setActive(log);
        LogFactory.getLog(SynchronisedAbstractionFactory.class).setActive(log);
        LogFactory.getLog(PrototypicalAbstractionFactory.class).setActive(log);
        // Sequence factories...
        LogFactory.getLog(StandardFactory.class).setActive(log);
        LogFactory.getLog(PrototypicalSequenceFactory.class).setActive(log);
        // Generic factories...
        LogFactory.getLog(TypeLiteral.class).setActive(true);
        LogFactory.getLog(Factory.class).setActive(log);
        LogFactory.getLog(TypedFactory.class).setActive(log);
        // Registry
        LogFactory.getLog(PrototypicalRegistry.class).setActive(log);
        
        //
        // Generator factories
        //

        // Prototypical registry...
        PrototypicalRegistry pr = new PrototypicalRegistry();
        
        out.heading("Generator factories");
        
        // Generator factories...
        CollectionValueFactory<String> cvfs = 
            new CollectionValueFactory<String>(String.class, add(new ArrayList<String>(), "1", "2", "2", "1", "0")); 
        CollectionValueFactory<Integer> cvfi = 
            new CollectionValueFactory<Integer>(Integer.class, add(new HashSet<Integer>(), 42, 117)); 
        CollectionValueFactory<Integer> cvfii = 
            new CollectionValueFactory<Integer>(Integer.class, add(new LinkedHashSet<Integer>(), 42, 66, 900, 32)); 
        RangeValueFactory rvf = new RangeValueFactory(1, 5); // Integer
        
        testGeneratorFactory(out, "1/4", cvfs,  add(new LinkedHashSet<String>(), "1", "2", "2", "1", "0"));
        testGeneratorFactory(out, "2/4", cvfi,  add(new ArrayList<Integer>(), 5, 4, 3, 2, 1));
        testGeneratorFactory(out, "3/4", cvfii, add(new TreeSet<Integer>(), 5, 4, 3, 2, 1));
        testGeneratorFactory(out, "4/4", rvf,   new Integer[]{10,100});
        
        
        //
        // Prototypes for prototypical factories
        //

        out.heading("Prototypical objects used for prototypical factories");
        
        // Abstractions...
        MemorizableSequenceAbstraction<String> msas = new MemorizableSequenceAbstraction<String>();
        SynchronisedSequenceAbstraction<Integer> ssai = new SynchronisedSequenceAbstraction<Integer>();
        // Already initialised..
        SequenceAbstraction<Integer> sai = 
            new SequenceAbstraction<Integer>(rvf.createGenerator(new Integer[]{1,10}));

        // 20'th Fibonacci number...
        BigInteger fibonacci20 = MathFacade.getInstance().getNthFibonnaciNumber(20);

        // Sequence...
        Sequence<String> ss = new SequenceAdapter<BigInteger,String>(
                                      new FibonacciSequence(fibonacci20),
                                      AdapterStrategy.ANY_TO_STRING
                                  );

        out.println("Prototype 1/4: ", msas)
           .println("Prototype 2/4: ", ssai)
           .println("Prototype 3/4: ", sai, " (initialised)")
           .println("Prototype 4/4: ", ss.toString(AdaptedPolicy.HIDDEN))
           .println();

        // Register prototypes to registry...
        pr.registerPrototype(MemorizableSequenceAbstraction.class, msas);
        pr.registerPrototype(SynchronisedSequenceAbstraction.class, ssai);
        pr.registerPrototype(FibonacciSequence.class, new FibonacciSequence(fibonacci20));
        // Non generic...
        pr.registerPrototype(LongSequence.class, new LongSequence(42));
        
        //
        // Abstraction factories
        //

        out.heading("Abstraction factories");
        
        StandardAbstractionFactory<String> safs = new StandardAbstractionFactory<String>();
        MemorizableAbstractionFactory<Integer> mafi = new MemorizableAbstractionFactory<Integer>();
        SynchronisedAbstractionFactory<String> safss = new SynchronisedAbstractionFactory<String>();
        PrototypicalAbstractionFactory<String> pafs = 
            new PrototypicalAbstractionFactory<String>(msas);
        PrototypicalAbstractionFactory<Integer> pafi = 
            new PrototypicalAbstractionFactory<Integer>(ssai);
        PrototypicalAbstractionFactory<Integer> pafii = 
            new PrototypicalAbstractionFactory<Integer>(sai);

        testAbstractionFactory(out, "1/6", safs);
        testAbstractionFactory(out, "2/6", mafi);
        testAbstractionFactory(out, "3/6", safss);
        testAbstractionFactory(out, "4/6", pafs);
        testAbstractionFactory(out, "5/6", pafi);
        testAbstractionFactory(out, "6/6", pafii);
        
        //
        // Sequence factories
        //

        out.heading("Abstraction factories");
        
        StandardFactory<String,Collection<String>> sfs = new StandardFactory<String,Collection<String>>();
        sfs.setAbstractionFactory(safss);
        sfs.setGeneratorFactory(cvfs);
        
        StandardFactory<Integer,Collection<Integer>> sfi = new StandardFactory<Integer,Collection<Integer>>();
        sfi.setAbstractionFactory(mafi);
        sfi.setGeneratorFactory(cvfi);

        StandardFactory<Integer,Integer[]> sfii = new StandardFactory<Integer,Integer[]>();
        sfii.setAbstractionFactory(pafi);
        sfii.setGeneratorFactory(rvf);
        
        PrototypicalSequenceFactory<String> psf = new PrototypicalSequenceFactory<String>(ss);
        
        testSequenceFactory(out, "1/1", sfs,  null);
        testSequenceFactory(out, "2/2", sfs,  add(new TreeSet<String>(), "foo", "bar", "foobar"));
        testSequenceFactory(out, "3/3", sfi,  null);
        testSequenceFactory(out, "4/4", sfi,  add(new HashSet<Integer>(), 13, 11, 7, 5, 3, 2));
        testSequenceFactory(out, "5/5", sfii, null);
        testSequenceFactory(out, "6/6", sfii, new Integer[]{2,13});
        testSequenceFactory(out, "7/7", psf, null);
        
        //
        // Changing abstraction factories
        //
        
        out.heading("Changing abstraction factories at runtime");
        
        // Changing factories...
        sfs.setAbstractionFactory(pafs);
        sfi.setAbstractionFactory(pafi);
        sfii.setAbstractionFactory(pafii);
        
        testSequenceFactory(out, "1/1", sfs,  null);
        testSequenceFactory(out, "2/2", sfs,  add(new TreeSet<String>(), "foo", "bar", "foobar"));
        testSequenceFactory(out, "3/3", sfi,  null);
        testSequenceFactory(out, "4/4", sfi,  add(new HashSet<Integer>(), 13, 11, 7, 5, 3, 2));
        testSequenceFactory(out, "5/5", sfii, null);
        testSequenceFactory(out, "6/6", sfii, new Integer[]{2,13});
        
        //
        // Changing generator factories
        //
        
        out.heading("Changing generator factories at runtime");

        sfi.setGeneratorFactory(cvfii);
        testSequenceFactory(out, "1/2", sfi,  null);
        testSequenceFactory(out, "2/2", sfi,  add(new HashSet<Integer>(), 13, 11, 7, 5, 3, 2));

        //
        // Testing prototypical registry
        //

        out.heading("Using prototypical registry");
        
        for (Class<? extends StrictCopyable<?>> type : pr.getRegisteredTypes()) {
            out.println("Creating product from registry: ", type.getName(), " -> ", pr.create(type));
        }
        out.println();
        
        //
        // Testing generic factories
        //

        out.heading("Using generic factories");

        try {
            
            // Generator factory...
            
            // 2 x Integer.class = minimum and maximum...
            TypedFactory<SequenceValueRange,Integer> fsvr = 
                new TypedFactory<SequenceValueRange,Integer>(Integer.class, 2){/*magic*/};

            // Abstraction factory...
            Factory<MemorizableSequenceAbstraction<Integer>> fmsai =
                new Factory<MemorizableSequenceAbstraction<Integer>>(){/*magic*/};
                
            // Sequence factory...
            TypedFactory<FibonacciSequence,BigInteger> ffs = 
                new TypedFactory<FibonacciSequence,BigInteger>(BigInteger.class){/*magic*/};
                
            out.println("Generic generator factory:   ", fsvr);
            out.println("Generic abstraction factory: ", fmsai);
            out.println("Generic sequence factory:    ", ffs);
            out.println();

            // Create implementation...
            SequenceValueRange svr = fsvr.createInstance(100, 105);
            // Create abstraction...
            MemorizableSequenceAbstraction<Integer> msai = fmsai.newInstance();
            // Attach generator for usage...
            msai.setGenerator(svr);
            
            // Create Fibonacci sequences...
            FibonacciSequence fs1 = ffs.createInstance(fibonacci20);
            FibonacciSequence fs2 = ffs.createInstance(fibonacci20.add(fibonacci20).add(BigInteger.ONE).add(BigInteger.ONE));
            FibonacciSequence fs3 = ffs.createInstance((BigInteger)null); // also allowed!

            // Compile time errors if commented in...
            // FibonacciSequence fs4 = ffs.createInstance(100);
            // FibonacciSequence fs5 = ffs.createInstance(new Date());
            // etc...
 
            out.println();
            out.println("Created generator:   ", svr);
            out.println("Created abstraction: ", msai);
            out.println("Created sequence 1:  ", fs1);
            out.println("Created sequence 2:  ", fs2);
            out.println("Created sequence 3:  ", fs3);
            
        } catch (Exception e) {
            out.error("Generic factory failure").error(e);
            ok = false;
        }
        
        out.heading("Abstract Factory tests ", ok ? "succeeded!" : "failed!");
        
        return ok;
    }

    /**
      * Test generator factories.
      * 
      * @param <E>      The type of values generated by the created generator.
      * @param <P>      The type of argument to use in the creation process. 
      * @param out      Log to use; never null.
      * @param index    The number of the test; never null.
      * @param factory  The factory to test; never null.
      * @param argument Argument used in the creation process, if any; can be null.
      */
    private final static <E,P> void testGeneratorFactory(Log out, String index, GeneratorFactory<E,P> factory, P argument) {
        P defaultArgument = factory.getDefaultArgument();
        out.println("Testing generator factory ", index, ": ", factory)
           .println("Default:  ", defaultArgument.getClass().getSimpleName(), defaultArgument, " -> ", factory.createGenerator())
           .println("Argument: ", argument.getClass().getSimpleName(),        argument,        " -> ", factory.createGenerator(argument))
           .println();
    }

    /**
      * Test abstraction factories.
      * 
      * @param <E>      The type of values generated by the created abstraction.
      * @param out      Log to use; never null.
      * @param index    The number of the test; never null.
      * @param factory  The factory to test; never null.
      */
    private final static <E> void testAbstractionFactory(Log out, String index, AbstractionFactory<E> factory) {
        out.println("Testing abstraction factory ", index, ": ", factory)
           .println("Created: ", factory.createAbstraction())
           .println();
    }

    /**
      * Test sequence factories.
      * 
      * @param  <E>      The type of values generated by the created sequence.
      * @param  <P>      The type of argument to use in the creation process. 
      * @param  out      Log to use; never null.
      * @param  index    The number of the test; never null.
      * @param  factory  The factory to test; never null.
      * @param  argument Argument used in the creation process, if any; can be null.
      * @return The created sequence; never null.
      */
    private final static <E,P> Sequence<E> testSequenceFactory(Log out, String index, SequenceFactory<E,P> factory, P argument) {
        Sequence<E> sequence = null;
        out.println("Testing sequence factory ", index, ": ", factory)
           .println("Created: ", (argument != null ? argument.getClass().getSimpleName() + argument : "null"), " -> ", sequence = factory.createSequence(argument))
           .print("Iterating over sequence values: ");
        
        boolean any = false;
        for (E value : new IterableSequence<E>(sequence)) {
            if (any) {
                out.print(", ");
            } else {
                any = true;
            }
            out.print(value);
        }
        out.println().println();
        return sequence;
    }
    
}
