/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.abstractfactory;

import dk.rode.thesis.bridge.SequenceAbstraction;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.ParticipantUsage;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>prototypical abstraction factory</i> creates 
  * sequence {@link SequenceAbstraction abstractions} based on
  * a prototype supplied at construction time. <p>
  * 
  * Created abstractions will be {@link SequenceAbstraction#isInitialised() 
  * initialised} if the prototype supplied at construction time was
  * initialised. <p>
  * 
  * This factory is thread-safe. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    MemorizableAbstractionFactory
  * @see    SynchronisedAbstractionFactory
  * 
  * @param  <E> The type of values delivered by the created abstractions.
  */
@ParticipantUsage(value = "ConcreteProduct", type = SequenceAbstraction.class)
@Participant("ConcreteFactory")
public class PrototypicalAbstractionFactory<E> implements AbstractionFactory<E>  {

    /**
      * The prototypical sequence abstraction to use. <p>
      * 
      * The abstraction may or may not have been initialised. <p>
      * 
      * Never null.
      */
    private final SequenceAbstraction<E> prototype;
    
    /**
      * Constructor. <p>
      * 
      * The abstraction supplied as {@code prototype} is copied, so
      * modifying it from an external context will not affect this
      * factory (unless {@code prototype} is a singleton). This factory
      * will not change {@code prototype}.
      * 
      * @param  prototype The prototypical abstraction to use; cannot be null.
      * @throws NullPointerException If {@code prototype} is null.
      */
    public PrototypicalAbstractionFactory(SequenceAbstraction<E> prototype) {
        // Defensive copy, so external contexts cannot alter the prototype...
        this.prototype = prototype.copy();
        // Reset sequence if already initialised so all copies will start from
        // the beginning of the sequence...
        if (this.prototype.isInitialised()) {
            this.prototype.reset();
        }
    }
    
    /**
      * 
      */
    public SequenceAbstraction<E> createAbstraction() {
        SequenceAbstraction<E> abstraction = this.prototype.copy(); // new instance may be initialised already!
        LogFactory.getLog(this).println("Created", (abstraction.isInitialised() ? " initialised " : " "), "abstraction from prototype: ", abstraction);
        return abstraction;
    }

    /**
      * Returns the string representation of this factory.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; prototype: ").append(this.prototype).append(']').toString();
    }
    
}
