/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.abstractfactory;

import dk.rode.thesis.factorymethod.Factory;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.ParticipantUsage;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.prototype.Copyable;
import dk.rode.thesis.prototype.StrictCopyable;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>prototypical factory</i> creates instances of a specific type of 
  * {@link Copyable copyable} type based on a prototype supplied at 
  * construction time. <p>
  * 
  * This factory is thread-safe. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * Using the {@linkplain dk.rode.thesis.prototype Prototype} pattern as a means
  * to create different types of sequences is an, perhaps safer, alternative 
  * to creating sequences using reflection. The {@link Factory} type illustrates
  * how reflection can be used in the creation process.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    PrototypicalRegistry
  * 
  * @param  <T> The type of values created by this factory.
  */
@ParticipantUsage(value = "ConcreteProduct", type = StrictCopyable.class)
@Participant("ConcreteFactory")
public class PrototypicalFactory<T extends StrictCopyable<T>> {

    /**
      * The prototypical object to use. <p>
      * 
      * Never null.
      */
    private final T prototype;
    
    /**
      * Constructor. <p>
      * 
      * {@code prototype} is copied and this factory keeps no reference
      * to it.
      * 
      * @param  prototype The prototypical object to use; cannot be null.
      * @throws NullPointerException If {@code prototype} is null.
      */
    public PrototypicalFactory(T prototype) {
        // Defensive copy, so external contexts cannot alter the prototype...
        this.prototype = prototype.copy();
    }
    
    /**
      * Creates a new instance based on the prototypical object supplied
      * at construction time. 
      * 
      * @return The new instance; never null.
      */
    public T create() {
        T object = this.prototype.copy();
        LogFactory.getLog(this).println("Created object from prototype: ", object);
        return object;
    }
    
    /**
      * Returns the string representation of this factory.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; prototype: ").append(this.prototype).append(']').toString();
    }
    
}
