/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.abstractfactory;

import dk.rode.thesis.bridge.SequenceAbstraction;
import dk.rode.thesis.bridge.SequenceValueGenerator;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>standard factory</i> can create sequence 
  * {@link SequenceAbstraction abstractions} as well as the sequence 
  * value {@link SequenceValueGenerator generators} used as the 
  * implementations for the abstractions, and further more provides
  * a convenient {@link #createSequence(Object) creation method} to 
  * create sequences in a single step. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * This factory works by composition: attached {@link AbstractionFactory} and
  * {@link GeneratorFactory} instances perform the actual creation, and
  * they can be changed at runtime, making this factory very flexible. 
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    PrototypicalSequenceFactory
  * @see    SequenceFactory
  * @see    AbstractionFactory
  * @see    GeneratorFactory 
  * 
  * @param  <E> The type of values delivered by the created abstractions and 
  *             generators.
  * @param  <P> The type of argument required in the creation process of
  *             generators. Use {@code Void} if no arguments are required.
  */
@Participant("ConcreteFactory")
public class StandardFactory<E,P> 
  implements SequenceFactory<E,P>, AbstractionFactory<E>, GeneratorFactory<E,P> {

    /**
      * The currently used abstraction factory. <p>
      * 
      * The factory may change at runtime. <p>
      * 
      * Will be null until explicitly set. 
      */
    private AbstractionFactory<E> abstractionFactory;
    /**
      * The currently used generator factory. <p>
      * 
      * The factory may change at runtime. <p>
      * 
      * Will be null until explicitly set. 
      */
    private GeneratorFactory<E,P> generatorFactory;
  
    /**
      * Constructor. <p>
      * 
      * {@link AbstractionFactory Abstraction} and {@link GeneratorFactory generator}
      * factories must be explicit set before this factory can be used to create
      * abstractions and generators, respectively.
      * 
      * @see #setAbstractionFactory(AbstractionFactory)
      * @see #setGeneratorFactory(GeneratorFactory)
      */
    public StandardFactory() {
        super();
    }
    
    /**
      * Constructor. <p>
      * 
      * The {@link AbstractionFactory abstraction} and {@link GeneratorFactory generator}
      * factories supplied as {@code abstractionFactory} and 
      * {@code generatorFactory} will be set as the factories used,
      * respectively. They can be changed at runtime if so desired.
      * 
      * @param  abstractionFactory The abstraction factory to use; cannot be null.
      * @param  generatorFactory   The generator factory to use; cannot be null.
      * @throws NullPointerException If either argument is null.
      * @see    #setAbstractionFactory(AbstractionFactory)
      * @see    #setGeneratorFactory(GeneratorFactory)
      */
    public StandardFactory(AbstractionFactory<E> abstractionFactory, GeneratorFactory<E,P> generatorFactory) {
        this.setAbstractionFactory(abstractionFactory);
        this.setGeneratorFactory(generatorFactory);
    }

    /**
      * Returns the {@link AbstractionFactory abstraction} factory
      * currently used by this factory. <p>
      * 
      * @return The factory; can be null if not set.
      */
    public AbstractionFactory<E> getAbstractionFactory() {
        return this.abstractionFactory;
    }
    
    /**
      * Sets the {@link AbstractionFactory abstraction} factory
      * to use by this factory to {@code abstractionFactory}. <p>
      * 
      * Any previously set abstraction factory will be discarded,
      * if any. 
      * 
      * @param  abstractionFactory The factory; cannot be null.
      * @throws NullPointerException If {@code abstractionFactory} is null.
      */
    public void setAbstractionFactory(AbstractionFactory<E> abstractionFactory) {
        if (abstractionFactory == null) {
            throw new NullPointerException("Factory cannot be null");
        }
        this.abstractionFactory = abstractionFactory;
    }

    /**
      * Returns the {@link GeneratorFactory generator} factory
      * currently used by this factory. <p>
      * 
      * @return The factory; can be null if not set.
      */
    public GeneratorFactory<E,P> getGeneratorFactory() {
        return this.generatorFactory;
    }
    
    /**
      * Sets the {@link GeneratorFactory generator} factory
      * to use by this factory to {@code generatorFactory}. <p>
      * 
      * Any previously set generator factory will be discarded,
      * if any. 
      * 
      * @param  generatorFactory The factory; cannot be null.
      * @throws NullPointerException If {@code generatorFactory} is null.
      */
    public void setGeneratorFactory(GeneratorFactory<E,P> generatorFactory) {
        if (generatorFactory == null) {
            throw new NullPointerException("Factory cannot be null");
        }
        this.generatorFactory = generatorFactory;
    }

    /**
      * 
      */
    public P getDefaultArgument() {
        return this.generatorFactory.getDefaultArgument();
    }
    
    /**
      * 
      */
    public SequenceValueGenerator<E> createGenerator() {
        return this.createGenerator(this.getDefaultArgument());
    }
    
    /**
      * 
      */
    public SequenceValueGenerator<E> createGenerator(P argument) {
        // Can throw a null pointer if no factory set...
        return this.generatorFactory.createGenerator(argument);
    }

    /**
      * 
      */
    public SequenceAbstraction<E> createAbstraction() {
        // Can throw a null pointer if no factory set...
        return this.abstractionFactory.createAbstraction();
    }
    
    /**
      * {@inheritDoc}
      * 
      * First, a {@link SequenceAbstraction} is created by the 
      * abstraction {@link #getAbstractionFactory() factory} currently
      * used. If the abstraction is {@link SequenceAbstraction#isInitialised() 
      * initialised}, it will be returned. If not, a {@link SequenceValueGenerator
      * generator} will be created using the argument supplied as {@code argument},
      * if any, by the generator {@link #getGeneratorFactory() factory} currently
      * used. If no argument is supplied, the {@link #getDefaultArgument() default}
      * argument is used. Then the generator is attached as the implementation for the
      * created abstraction, which is finally returned.
      * 
      * @param  argument A context supplied argument to use in creation process. If
      *                  the created abstraction is initialised, the argument is 
      *                  unused. Can be null, in which case the default argument
      *                  is used.
      */
    public Sequence<E> createSequence(P argument) {
        // Creator abstraction...
        SequenceAbstraction<E> abstraction = this.createAbstraction();
        // May already have been initialised, for example if based on
        // a prototype...
        if (abstraction.isInitialised()) {
            LogFactory.getLog(this).println("Created (already initialised) sequence: ", abstraction);
            return abstraction;
        }
        // Create implementation...
        SequenceValueGenerator<? extends E> generator = this.createGenerator(argument);
        // Initialise and return...
        Sequence<E> sequence = abstraction.setGenerator(generator);
        LogFactory.getLog(this).println("Created sequence: ", sequence);
        return sequence;
    }
    
    /**
      * Returns the string representation of this factory.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; abstraction-factory: ").append(this.abstractionFactory)
                 .append("; generator-factory: ").append(this.generatorFactory).append(']').toString();
    }
    
}
