/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */

/**
  * Implementations and examples of the <b>Abstract Factory</b> 
  * design pattern [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.87]. <p>
  *  
  * <b>Intent</b>:<ul><p><i>
  * 
  *    Provide an interface for creating families of related or dependent 
  *    objects without specifying their concrete classes.
  *     
  * </i></ul>
  *
  * Here, the <b>AbstractProduct</b> participant is represented by the
  * {@link dk.rode.thesis.bridge.SequenceAbstraction},
  * {@link dk.rode.thesis.bridge.SequenceValueGenerator}, and 
  * {@link dk.rode.thesis.meta.model.Sequence} types. Hence, the
  * <b>ConcreteProduct</b> participant is represented by any class
  * inheriting or implementing the aforementioned types, including
  * {@code SequenceAbstraction} itself. <p>
  * 
  * A factory type is defined to create each of the abstract product
  * types, respectively the {@link dk.rode.thesis.abstractfactory.AbstractionFactory},
  * {@link dk.rode.thesis.abstractfactory.GeneratorFactory}, and 
  * {@link dk.rode.thesis.abstractfactory.SequenceFactory} interfaces.  
  * All theses factories represent the <b>AbstractFactory</b> participant. <p>
  * 
  * Several different factory implementations have been made to 
  * represent the <b>ConcreteFactory</b> participant, each capable of
  * creating one or more of described product types. The 
  * {@link dk.rode.thesis.abstractfactory.StandardFactory} implements
  * all three factory types, but uses composition to let 
  * {@code AbstractionFactory} and {@code GeneratorFactory} instances
  * perform the actual creation; sequences created by it are simply 
  * initialised {@code SequenceAbstraction} instances. Specific 
  * abstraction factories include 
  * {@link dk.rode.thesis.abstractfactory.StandardAbstractionFactory} and
  * {@link dk.rode.thesis.abstractfactory.MemorizableAbstractionFactory}, while
  * {@link dk.rode.thesis.abstractfactory.CollectionValueFactory} and
  * {@link dk.rode.thesis.abstractfactory.RangeValueFactory} are the only two
  * generator factories. <p> 
  *   
  * Several prototypical factories are also provided, that create products based
  * on prototypes. The {@link dk.rode.thesis.abstractfactory.PrototypicalFactory}
  * is a stand-alone prototypical factory, while the {@link dk.rode.thesis.abstractfactory.PrototypicalRegistry}
  * class is a registry that can create multiple product types. <p>
  *    
  * Finally, the <b>Client</b> participant is represented by the
  * {@link dk.rode.thesis.abstractfactory.Main} class, which creates
  * and uses the created sequences. <p>
  * 
  * <a name="uml"></a><b>UML Class Diagram</b>:<p>
  * <img src="doc-files/AbstractFactory.png" border="0"><p>
  * 
  * <b>Implementation notes</b>:<br>
  * The {@link dk.rode.thesis.abstractfactory.PrototypicalSequenceFactory} factory
  * illustrates how the {@linkplain dk.rode.thesis.prototype Prototype} pattern can
  * be used to create objects based on prototypes. The {@code PrototypicalSequenceFactory}
  * can create any type of sequence based on a prototype supplied at construction time.
  * The {@link dk.rode.thesis.abstractfactory.PrototypicalAbstractionFactory} provides
  * similar functionality, but for {@code SequenceAbstraction} instances. <p>
  * 
  * The use of prototypes is a possible alternative to using reflection to create different types
  * of products. However, the {@link dk.rode.thesis.abstractfactory.PrototypicalRegistry} 
  * illustrates that using a registry to store prototypes to create can cause problems with
  * type safety when generic types are created. The generic 
  * {@link dk.rode.thesis.factorymethod.Factory} class illustrates how reflection can be used 
  * to make a completely generic factory type to create any type of instantiable product, 
  * providing the product type has a usable constructor. Ideally, it should be used in 
  * an Abstract Factory implementation via composition, but here we simply illustrate its
  * use directly in {@code Main} and combine products as need be. 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a> 
  */
@Pattern(name         = "Abstract Factory", 
         scope        = Scope.Object,
         purpose      = Purpose.Creational,
         participants = {"AbstractFactory", "ConcreteFactory", "AbstractProduct", "ConcreteProduct", "Client"}) 
package dk.rode.thesis.abstractfactory;

import dk.rode.thesis.meta.Pattern;
import dk.rode.thesis.meta.Purpose;
import dk.rode.thesis.meta.Scope;






