/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.adapter;

import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.ObjectPolicy;
import dk.rode.thesis.strategy.SequencePolicy;
import dk.rode.thesis.strategy.StringablePolicy;

/**
  * Additional policies for formatting {@link SequenceAdapter} 
  * objects into char sequences (<b>not</b> part of the core Adapter 
  * implementation). <p>
  *
  * <b>Implementation notes</b>:<br>
  * <b>This is NOT part of the core Adapter functionality, merely 
  * to provide better log descriptions!</b><p>
  * 
  * See implementation notes for the {@link ObjectPolicy} enumeration.
  *    
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    SequencePolicy
  */
public enum AdaptedPolicy implements StringablePolicy<Sequence<?>> {

    /**
      * Policy used to format string representations of
      * adapters that indicates the nested structure. <p>
      * 
      * Note, however, that the check for adapted sequences
      * is done using {@code instanceof}, which will not discover
      * decorated or proxied sequences! <p>
      * 
      * Adapters using {@link #HIDDEN} as their default
      * policy are respected. 
      */
    VISIBLE {
       public CharSequence toString(Sequence<?> sequence) {
           StringBuilder sb = new StringBuilder();
           if (sequence instanceof SequenceAdapter) {
               SequenceAdapter<?,?> sd = (SequenceAdapter<?,?>)sequence;
               sb.append(sd.toString(SequencePolicy.NAME))
                 .append("[adapter: ")
                 .append(sd.getAdapterDelegate())
                 .append("].")
                 // Use default to respect hidden policies as well!
                 .append(sd.sequence.toString(SequencePolicy.DEFAULT)); // recurse...
           } else {
               sb.append(sequence.toString(SequencePolicy.VERBOSE));
           }
           return sb.toString();
       }
    }, 
    
    /**
      * Policy used to format string representations of
      * adapters that hides the nested structure "from
      * the sequence on". 
      */
    HIDDEN {
        public CharSequence toString(Sequence<?> sequence) {
            while (sequence instanceof SequenceAdapter) {
                sequence = ((SequenceAdapter<?,?>)sequence).sequence;
            }
            return sequence.toString(SequencePolicy.DEFAULT);
        }
    };

    /**
      * {@inheritDoc} 
      */
    public abstract CharSequence toString(Sequence<?> sequence);
    
    /**
      * @return {@link dk.rode.thesis.strategy.StringablePolicy.Type#APPENDABLE}. 
      */
    public Type type() {
        return Type.APPENDABLE;
    }
        
}
