/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.bridge;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.util.Const;
import dk.rode.thesis.prototype.Copyable;
import dk.rode.thesis.strategy.Stringable;

/**
  * A <i>sequence value generator</i> can be used as the <i>implementation</i>
  * for a sequence {@link SequenceAbstraction abstraction} as it generates 
  * non-null sequence values that can be delivered by the sequence in 
  * question. <p>
  * 
  * Any generator has the following properties: <ul><p>
  * 
  * <li> At least one value and at most {@link #size() size} values
  *      will be generated, possibly including duplicates, and 
  *      values are never null; <p>
  * 
  * <li> The generator itself is {@link #copy() copyable}, and
  *      will perform a <i>deep copy</i> if necessary, i.e.
  *      including generated values stored internally; <p>
  * 
  * <li> Can generate {@link #ordered() ordered} and/or 
  *      {@link #sorted() sorted} values, and may or may not generate 
  *      {@link #duplicates() duplicate} values; <p> 
  * 
  * <li> Keeps track of the number of generated values between
  *      calls to {@link #first()}; <p>
  * 
  * <li> Will generate a value on each call to {@link #get()}  
  *      or {@link #get(int)} adhering to the properties listed in the above 
  *      bullet, and will automatically restart when values are depleted; and <p>
  *      
  * <li> Can be reset to generate values <i>from the start</i> again by 
  *      calling {@link #first()}. <p> 
  * 
  * </ul>
  * 
  * Note, that the {@code size}, {@code ordered}, and {@code duplicate} 
  * properties makes it <b>very</b> easy to alter the behaviour of a given
  * sequence {@link SequenceAbstraction abstraction}! <p> 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values generated by this generator.
  */
@Participant("Implementor")
public interface SequenceValueGenerator<E> 
         extends Copyable<SequenceValueGenerator<E>>, 
                 Stringable<SequenceValueGenerator<E>> {

    /**
      * Returns the actual type of the values generated by this generator. 
      *  
      * @return The value type; never null.
      */
    public Class<E> getValueType();

    /**
      * Returns the number of values that can be generated by this
      * generator in total, excluding re-generation of values 
      * after a call to {@link #first()}. <p>
      *
      * Whether or not the count include {@link #duplicates() duplicates}
      * is sub-class specific. If {@code duplicates} is false, the
      * size is always equal to the number of unique values generated
      * by this generator.
      * 
      * @return The number of values, or {@link Const#INFINITE} if this
      *         generator has no (practical) bounds on the number of 
      *         values it can generate.
      */
    public int size();
    
    /**
      * Returns the number of the the last value generated by
      * either {@link #first()}, {@link #get()}, or {@link #get(int)}. <p>
      * 
      * The number of generated values is reset when {@code first()}
      * is invoked, i.e. to a number of one. <p>
      * 
      * Notice, that {@code first} and {@code get(int)} alter the 
      * number count of values generated, which will be reflected in 
      * the next invocation of {@code get()}! <p> 
      * 
      * Range: 1 (0) < number < {@link #size()}, if {@code size != INFINITE}. <p>
      * 
      * A number of zero will only occur when no values have been 
      * generated yet.
      * 
      * @return The number of the last value generated.
      */
    public int number();
    
    /**
      * Resets this generator to re-generate its <i>first</i> value, 
      * where after values can (again) be generated by {@link #get()}. <p>
      * 
      * The {@link #number() number} of generated values is reset when 
      * this method is invoked. <p>
      * 
      * If this generator does not generate {@link #ordered() ordered}, nor
      * {@link #sorted() sorted} values, it is undefined which value is 
      * generated as the first value. 
      *  
      * @return A <i>first</i> value generated by this generator; never null.
      */
    public E first();
    
    /**
      * Returns a value generated by this generator. <p>
      * 
      * The number of the generated value is one higher than the last 
      * value returned by either {@link #first()}, {@link #get()},
      * or {@link #get(int)}. <p>
      * 
      * The {@link #number() number} of generated values is reset 
      * when {@code first()} is invoked. <p>
      * 
      * If this generator generates {@link #ordered() ordered} or 
      * {@link #sorted() sorted} values, the generation of values 
      * is consistent even after {@link #first()} have been invoked. 
      * Otherwise the order is undefined. 
      * 
      * @return A value generated by this generator; never null.
      * @see    #get(int)
      */
    public E get();
    
    /**
      * Returns the {@code number}'th value generated by this generator,
      * which may or may not have been generated yet. The next call to 
      * {@link #get()} from generate and return the {@code number + 1}'th 
      * value. <p>
      * 
      * The {@link #number() number} of generated values is reset 
      * when {@code first()} is invoked. <p>
      * 
      * If this generator generates {@link #ordered() ordered} or 
      * {@link #sorted() sorted} values, the generation of values is 
      * consistent even after {@link #first()} have been invoked. Otherwise
      * the order is undefined, and this method may generate
      * different values for the same number each time invoked.
      * 
      * @param  number The number in question; cannot be zero or negative.
      *                Is one-, not zero-based!
      * @return The {@code number}'th value generated by this generator; 
      *         never null.
      * @throws IllegalArgumentException If {@code number} is zero or negative 
      *         or larger than the {@link #size() maximum} number of values 
      *         that can be generated by this generator, if a maximum is 
      *         enforced.
      * @see    #get()
      * @see    #numberOf(Object, int)
      */
    public E get(int number);
    
    /**
      * Returns the number of the generated value supplied as {@code value}
      * using an offset of {@code fromNumber}, or {@link Const#EOF} if no 
      * such value will ever been generated by this generator. <p> 
      *
      * Unlike {@link #first()}, {@link #get()}, and {@link #get(int)}, this 
      * method will <b>not</b> alter the number of values generated by this 
      * generator. <p>
      * 
      * @param  value      The value to test; cannot be null.
      * @param  fromNumber The number to start from; only relevant if the
      *                    values generated contain {@link #duplicates()
      *                    duplicates}. Is one-, not zero-based! If
      *                    larger than {@link #size()}, this method will
      *                    always return {@code EOF}.
      * @return The number of the generated value supplied as {@code value}, 
      *         or {@code EOF} if no such value can be generated by 
      *         this value generator.
      * @throws NullPointerException If {@code value} is null.
      * @throws IllegalArgumentException If {@code fromNumber} is zero or negative. 
      * @see    #get(int)
      */
    public int numberOf(E value, int fromNumber);
    
    /**
      * Returns true if the values generated by this generator
      * are ordered, i.e. always delivered in <i>some</i> known 
      * order.
      * 
      * @return True if ordered, false if not.
      */
    public boolean ordered();
    
    /**
      * Returns true if the values generated by this generator
      * are sorted, for example based on numerical value in case
      * of numbers. 
      *  
      * @return True if sorted, false if not.
      */
    public boolean sorted();
    
    /**
      * Returns true if this generator will generate duplicate (equivalent) 
      * values, false if not. <p>
      * 
      * The test for duplicates is performed using {@code equals(Object)}.
      * 
      * @return True if this generator will generate duplicate values, false
      *         if not.
      */
    public boolean duplicates();
 
    /**
      * Performs a <i>deep copy</i> of this generator. <p>
      *
      * The {@link #size() size}, {@link #ordered() ordered}, 
      * {@link #sorted() sorted}, and {@link #duplicates() duplicate} 
      * properties are adhered to by the returned clone. <p>
      *
      * @return A copy of this generator; never null.
      */
    public SequenceValueGenerator<E> copy();
        
}
