/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.bridge;

import dk.rode.thesis.meta.Participant;

/**
  * A <i>synchronised sequence abstraction</i> refines a sequence 
  * {@link SequenceAbstraction abstraction} by making it thread-safe. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by this sequence. 
  */
@Participant("RefinedAbstraction")
public class SynchronisedSequenceAbstraction<E> extends SequenceAbstraction<E> {

    /**
      * Constructor. <p>
      * 
      * A value {@link SequenceValueGenerator generator} <b>must</b> explicitly 
      * be set using the {@link #setGenerator(SequenceValueGenerator)} 
      * method before this sequence is valid!  
      */
    public SynchronisedSequenceAbstraction() {
        super();
    }
    
    /**
      * Constructor, which creates this synchronised sequence abstraction 
      * to use the value {@link SequenceValueGenerator generator}
      * supplied as {@code generator}. <p>
      *
      * The {@link #setGenerator(SequenceValueGenerator)} must <b>not</b>
      * be invoked for this sequence.
      * 
      * @param  generator The generator to use; cannot be null.
      * @throws NullPointerException If {@code generator} is null.
      */
    public SynchronisedSequenceAbstraction(@Participant("Implementation") SequenceValueGenerator<? extends E> generator) {
        super(generator);
    }

    /**
      * Copy constructor. <p>
      * 
      * If {@code sequence} does not have a value {@link SequenceValueGenerator 
      * generator} set, a generator <b>must</b> explicitly 
      * be set for this sequence using the 
      * {@link #setGenerator(SequenceValueGenerator)} method before 
      * this sequence is valid!  
      * 
      * @param  sequence The (synchronised) sequence abstraction to copy; 
      *                  cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public SynchronisedSequenceAbstraction(SequenceAbstraction<E> sequence) {
        super(sequence);
    }
    
    /**
      * 
      */
    public SynchronisedSequenceAbstraction<E> copy() {
        return new SynchronisedSequenceAbstraction<E>(this);
    }

    /**
      * 
      */
    public synchronized SequenceAbstraction<E> setGenerator(@Participant("Implementation") SequenceValueGenerator<? extends E> generator) 
        throws IllegalStateException {
        return super.setGenerator(generator);
    }
    
    /**
      * @throws IllegalStateException If a values has not been set.
      */
    public synchronized E current() throws IllegalStateException {
        return super.current();
    }
    
    /**
      * @throws IllegalStateException If a values has not been set.
      */
    public synchronized E next() throws IllegalStateException {
        return super.next();
    }

    /**
      * @throws IllegalStateException If a values has not been set.
      */
    public synchronized void reset() throws IllegalStateException {
        super.reset();
    }
    
}
