/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.builder;

import java.util.List;

import dk.rode.thesis.interpreter.Context;
import dk.rode.thesis.interpreter.Expression;
import dk.rode.thesis.interpreter.ExpressionException;
import dk.rode.thesis.interpreter.FlowExpression;
import dk.rode.thesis.interpreter.TypedExpression;

/**
  * A <i>typed flow expression</i> represents one or more expressions
  * to be evaluated in order, one at a time, where the {@link #type() type}
  * of value the evaluation of the expressions produces associated with it 
  * at runtime. The result of the evaluation is the result of the <i>last</i> 
  * expression evaluated.<p>
  *
  * Expressions to be evaluated by a flow expression must explicitly be
  * {@link #add(Expression) added}. Once the last expression has been 
  * added, the flow expression must be {@link #initialise() initialised} 
  * before evaluation, or the evaluation will fail. Once initialised, 
  * additional expressions cannot be added. <p> 
  *
  * <b>Implementation notes</b>:<br>
  * This is an application of the {@linkplain dk.rode.thesis.decorator Decorator}
  * and {@linkplain dk.rode.thesis.adapter Adapter} patterns. <p>
  * 
  * As {@code FlowExpression} returns a {@code FlowExpression} type from
  * several methods, these methods have to be overridden using covariant return
  * types as not to loose the type information. This requires creating a 
  * lot of instances of this class, all using the same decorated flow expression. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    TypedExpressionBuilder
  * 
  * @param  <E> The type of value the evaluation of this expression 
  *             produces.
  */
public class TypedFlowExpression<E> extends FlowExpression<E> implements TypedExpression<E> {

    /**
      * A class literal representing the type of value the evaluation 
      * of {@link #expression} produces. <p>
      * 
      * Never null.
      */
    private final Class<E> type;
    /**
      * The decorated flow expression. <p>
      * 
      * Never null, and never empty.
      */
    private final FlowExpression<E> expression;
    
    /**
      * Constructor. <p>
      *
      * @param  type       A class literal representing the type of value the 
      *                    evaluation of {@code expression} produces.
      * @param  expression The flow expression to use; cannot be null.
      * @throws NullPointerException If either argument is null.
      */
    public TypedFlowExpression(Class<E> type, FlowExpression<E> expression) {
        if ((this.type = type) == null) {
            throw new NullPointerException("Type cannot be null");
        } else if ((this.expression = expression) == null) {
            throw new NullPointerException("Expression cannot be null");
        }
    }

    /**
      * Copy constructor.
      * 
      * @param  expression The expression to copy; cannot be null.
      * @throws NullPointerException If {@code expression} is null.
      */
    public TypedFlowExpression(TypedFlowExpression<E> expression) {
        this.type = expression.type;
        this.expression = expression.expression.copy();
    }
 
    /**
      * 
      */
    public TypedFlowExpression<E> copy() {
        return new TypedFlowExpression<E>(this);
    }

    /**
      * 
      */
    public Class<E> type() {
        return this.type; // so much work for this...! :)
    }
    
    /**
      *
      */
    @Override
    public boolean isInitialised() {
        return this.expression.isInitialised();
    }

    /**
      * 
      */
    @Override
    public TypedFlowExpression<E> initialise() throws ExpressionException {
        return new TypedFlowExpression<E>(this.type, this.expression.initialise());
    }

    /**
      * 
      */
    @Override
    public TypedFlowExpression<E> add(Expression<? extends E> expression) throws ExpressionException {
        return new TypedFlowExpression<E>(this.type, this.expression.add(expression));
    }
    
    /**
      *  
      */
    @Override
    public List<Expression<?>> operands() {
        return this.expression.operands();
    }

    /**
      * 
      */
    @Override
    public E evaluate(Context context) throws ExpressionException {
        return this.expression.evaluate(context);
    }

    /**
      * 
      */
    @Override
    public String asSymbol(Context context) throws ExpressionException {
        return this.expression.asSymbol(context);
    }

    /**
      * Returns the hash code of this expression.
      * 
      * @return The hash code.
      */
    public int hashCode() {
        return this.type.hashCode() ^ this.expression.hashCode();
    }
    
    /**
      * Returns true if {@code object} is a type flow expression using
      * the same type and the same expression, false if not.
      * 
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object object) {
        if (!(object instanceof TypedFlowExpression)) {
            return false;
        }
        TypedFlowExpression<?> tfe = (TypedFlowExpression<?>)object;
        return this.type.equals(tfe.type) && this.expression.equals(tfe.expression);
    }
    
} 
