/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.chainofresponsibility;

import dk.rode.thesis.meta.Participant;

/**
  * A <i>handler</i> can handle a given type of request by taking
  * appropriate action <b>or</b> by forwarding the request to the 
  * next handler in the current chain of handlers. <p>
  * 
  * A handler is only {@link HandlerLink linked} to the next handler 
  * at execution time, and hence a handler can participate in several 
  * handler chains, which the original "Gang of Four" solution could 
  * not. <p>
  * 
  * The actual code required to support the handler chain semantics 
  * is minimal in handler implementations: 
  * <pre>
  *   public Handler&lt;R> {@link #handle(Object, HandlerLink) handle}(R request, {@link HandlerLink}&lt;R> link) {
  *     // Unique for each handler implementation...
  *     if (.. request can be handled ..) {
  *       ..  
  *       return this;
  *     }  
  *     // Identical for *every* handler implementation...
  *     return link == null ? null : link.{@link HandlerLink#forward(Object, Handler) handle}(request, this);  
  *   }
  *  
  * </pre>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <R> The request type handled by this handler.
  */
@Participant("Handler")
public interface Handler<R> {

    /**
      * Handles the request supplied as {@code request} or
      * forwards it to the next handler in the chain represented 
      * by the handler {@link HandlerLink link} supplied as 
      * {@code link}, if any. {@code link} can be null,
      * in which case this handler is considered the last
      * handler in the chain. <p>
      * 
      * If this handler handles {@code request}, this handler
      * is returned. If this handler cannot handle {@code request},
      * the result from handling the request using {@code link}
      * is returned. Finally, if {@code link} is null, null is 
      * returned. 
      * 
      * @param  request The request to handle; nullability determined
      *                 by the actual handler implementation.
      * @param  link    The link to the next handler in the current 
      *                 chain; can be null.
      * @return The handler that handled {@code request}, if any. 
      *         Can be null.
      */
    public Handler<R> handle(R request, HandlerLink<R> link);

}
