/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.chainofresponsibility;

import java.util.List;

import dk.rode.thesis.composite.CompositeSequence;
import dk.rode.thesis.meta.reflect.Executor;
import dk.rode.thesis.observer.ObservableSequence;
import dk.rode.thesis.observer.ObserverManager;

/**
  * A <i>handler chain</i> is an ordered list of {@link Handler handlers} 
  * that will be passed a <i>request</i> in turn until the first applicable
  * handler can handle it. To handle a given request, the 
  * {@link #handle(Object)} method is used. <p> 
  * 
  * Any handler can {@link #register(Handler) register}
  * it self to a proper handler chain, and <i>delegate</i> the chain 
  * handling responsibility to the chain in case the handler cannot 
  * handle the request. <p>
  * 
  * Handler chains can change at runtime. <p> 
  * 
  * <b>Implementation notes</b>:<br>
  * The handler chain is {@linkplain dk.rode.thesis.composite Composite} 
  * and allows other handler chains to be registered as handlers. <p>
  * 
  * An alternative to using handler interfaces is to use annotations
  * to mark handlers in a given class. The  
  * {@linkplain dk.rode.thesis.observer Observer} pattern utilises such 
  * an approach to identify the {@link Executor methods} (handlers) 
  * to notify relevant observers in the {@link ObserverManager} class. 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  *
  * @see    HandlerLink
  * @see    ObservableSequence
  * @see    CompositeSequence
  * 
  * @param  <R> The type of requests handled by this chain.
  */
public interface HandlerChain<R> extends Handler<R>, HandlerLink<R> {

    /**
      * Returns an ordered list of the {@link Handler handlers} 
      * currently associated with this chain, starting from
      * the handler <b>after</b> {@code handler}, if supplied. <p> 
      * 
      * If {@code handler} is null, the entire list of handlers
      * is returned. If {@code handler} is not registered to
      * this chain, the entire list of handlers is returned. <p>
      *
      * Modifying the returned list will not affect this chain. <p>
      * 
      * @param  handler The handler to start from; can be null.
      * @return The ordered list of handlers, starting <b>after</b> 
      *         {@code handler}; never null, but can be empty.
      */
    public List<Handler<R>> getHandlers(Handler<R> handler);
    
    /**
      * Returns true if {@code handler} is registered to this handler
      * chain, false if not.
      * 
      * @param  handler The handler to test; cannot be null.
      * @return True if registered, false if not.
      * @throws NullPointerException If {@code handler} is null.
      */
    public boolean isRegistered(Handler<R> handler);
    
    /**
      * Return the size of this handler chain. <p>
      * 
      * @return The size.
      */
    public int size();
    
    /**
      * Registers the {@link Handler handler} supplied
      * as {@code handler} to this handler chain. <p>
      * 
      * Registering the same handler twice is an error.
      * 
      * @param  handler The handler to add; cannot be null.
      * @throws NullPointerException If {@code handler} is null.
      * @throws IllegalArgumentException If {@code handler} is already
      *         registered to this chain.
      * @see    #unregister(Handler)         
      */
    public void register(Handler<R> handler);
    
    /**
      * Removes the {@link Handler handler} supplied
      * as {@code handler} from this handler chain. <p>
      * 
      * Trying to remove a handler that is not associated with
      * this handler chain is an error. 
      * 
      * @param  handler The handler to remove; cannot be null.
      * @throws NullPointerException If {@code handler} is null.
      * @throws IllegalArgumentException If {@code handler} is not
      *         registered to this chain.
      * @see    #register(Handler)         
      */
    public void unregister(Handler<R> handler);

    /**
      * Handles the request supplied as {@code request} by forwarding
      * it to each {@link Handler handler} in this chain until a 
      * handler can handle it, starting from the first registered handler. <p>
      * 
      * The handlers used are those delivered by {@link #getHandlers(Handler)}
      * with a null argument.
      * 
      * @param  request The request to handle. This chain allows it to 
      *                 be null, but individual handlers may not.
      * @return The handler that handled {@code request}, if any; may be null. 
      */
    public Handler<R> handle(R request);

}
