/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.command;

import java.util.List;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>null command</i> performs no operation on {@link #execute() 
  * execute} or on {@link #undo() undo}, is always considered
  * {@link #isUndoable() undoable}, and carries no state. <p>
  *
  * Null commands are thread-safe and can be shared. 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of result obtained by executing this command. 
  *             This command type always return null as the result.
  */
@Participant("ConcreteCommand")
public class NullCommand<E> implements Command<E> {
    
    /**
      * No-arg constructor.
      */
    public NullCommand() {
        super();
    }

    /**
      * 
      */
    public NullCommand<E> copy() {
        return new NullCommand<E>();
    }
    
    /**
      * Does nothing.
      * 
      * @return Always null.
      */
    public List<Command<E>> execute() {
        return null;
    }

    /**
      * Returns true.
      * 
      * @return True.
      */
    public boolean isUndoable() {
        return true;
    }

    /**
      * Does nothing.
      * 
      * @return Always null.
      */ 
    public E undo() {
        return null;
    }
    
    /**
      * @return Always null. 
      */
    public E getResult() {
        return null;
    }
    
    /**
      * Returns the hash code of this null command. 
      *
      * @return The hash code.
      */
    public int hashCode() {
        return this.getClass().hashCode();
    }
    
    /**
      * Returns true if {@code object} is an instance
      * of a {@code NullCommand}.
      *
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object object) {
        return (object instanceof NullCommand);
    }
    
    /**
      * Returns the string representation of this null command.
      *  
      * @return The string representation; never null.
      * @see    ObjectPolicy#ID
      */
    public String toString() {
        return ObjectPolicy.ID.asString(this);
    }
    
}
