/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.decorator;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.util.Strings;

/**
  * An <i>upper-case sequence decorator</i> returns the string value
  * returned by the decorated sequence in upper-case, for example 
  * {@code a -> A}. <p>
  * 
  * A flag can be supplied at construction time which specifies if
  * only the first character from the string value should be 
  * upper-cased, i.e. {@code aaa -> Aaa} as opposed to 
  * {@code aaa -> AAA} (default). <p>
  * 
  * Only sequences returning strings can be decorated. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("ConcreteDecorator")
public class UppercaseDecorator extends SequenceDecorator<String> implements Sequence<String> {

    /**
      * True if only the first character in the sequence value
      * should be upper-cased, false if the entire value should
      * be upper-cased. 
      */
    private final boolean firstOnly;
    
    /**
      * Constructor. <p>
      * 
      * The entire sequence value (String) will be upper-cased. 
      * 
      * @param  sequence The sequence to decorate; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public UppercaseDecorator(@Participant("ConcreteComponent") Sequence<String> sequence) {
        this(sequence, false);
    }
    
    /**
      * Constructor. <p>
      * 
      * If {@code firstOnly} is true, only the first character
      * in the values delivered by {@code sequence} will be
      * upper-cased.
      * 
      * @param  sequence  The sequence to decorate; cannot be null.
      * @param  firstOnly True if only the first character should
      *                   be upper-cased, false for the entire
      *                   value.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public UppercaseDecorator(@Participant("ConcreteComponent") Sequence<String> sequence, boolean firstOnly) {
        super(sequence);
        this.firstOnly = firstOnly;
    }
    
    /**
      * 
      */
    public UppercaseDecorator copy() {
        return new UppercaseDecorator(this.getSequence(true), this.firstOnly);
    }

    /**
      * 
      */
    public String current() {
        return toUpperCase(super.current(), this.firstOnly);
    }
    
    /**
      * 
      */
    public String next() {
        return toUpperCase(super.next(), this.firstOnly);
    }

    /**
      * Returns an upper-cased string based on {@code value}.
      *  
      * @param  value     The value to upper-case (first); never null.
      * @param  firstOnly True if only the first character in 
      *                   {@code value} should be upper-cased.
      * @return The upper-cased (first) string; never null.
      */
    private final static String toUpperCase(String value, boolean firstOnly) {
        if (firstOnly) {
            return Strings.upperCaseFirst(value).toString();
        } else {
            return value.toUpperCase();
        }
        
    }
}
