/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.facade;

import java.math.BigInteger;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import dk.rode.thesis.iterator.SequenceIterator;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.ParticipantUsage;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.state.ReversiblePrimeSequence;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>math facade</i> is a small math library that represents <i>an-easy-to-use</i>
  * high-level interface for complex mathematical functions. (yeah, right...) <p>
  * 
  * Facade usage is thread-safe. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * The small examples used throughout this thesis does not really
  * justify use of Facade; they are simply not complex enough. Still, 
  * this Facade implementation address several issues discussed by
  * Gamma et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.185-193]: <ul>
  * 
  * <li> The mathematical functions implemented with {@link Sequence sequences}
  *      are too verbose and tedious to be implemented by any client, and hence
  *      this facade help provide the desired functionality, shielding clients 
  *      from the gory details. <p>
  * 
  * <li> The facade shield the client from knowing about specific sub-system
  *      classes, i.e. sequences. <p>
  * 
  * <li> The facade implementation use package private access to ensure
  *      certain sub-system classes are off-limit from clients. <p> 
  * 
  * <li> The facade is implemented as a {@linkplain dk.rode.thesis.singleton Singleton}. <p>
  * 
  * </ul>
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("Facade")
public class MathFacade {

    /**
      * The singleton instance to use. <p>
      * 
      * Never null.
      */
    private final static MathFacade instance = new MathFacade();
    
    /**
      * The prime sequence to deliver the prime numbers. <p>
      * 
      * Will "grow in size" as needed. <p>
      * 
      * Lazily created.
      */
    @ParticipantUsage(value = "SubsystemClass", type = ReversiblePrimeSequence.class)
    private ReversiblePrimeSequence primes;
    /**
      * A map storing calculated Ackermann values, <i>A(m,n)</i>, 
      * per <i>m</i> value. <p>
      * 
      * Never null.
      */
    private Map<Integer,AckermannSequence> ackermann;
    /**
      * An unbounded random sequence that uses a Fibonacci sequence
      * as the seed. <p>
      * 
      * Never null.
      */
    private final UnboundedRandomSequence random;
    
    /**
      * Private constructor as this class is a singleton.
      */
    private MathFacade() {
        this.ackermann = new HashMap<Integer,AckermannSequence>();
        this.random = new UnboundedRandomSequence(
                          new FibonacciSequence(this.getNthFibonnaciNumber(100)) 
                       );
    }
    
    /**
      * Returns the math library to use. 
      *  
      * @return The math library; never null.
      */
    public final static MathFacade getInstance() {
        return instance;
    }

    /**
      * Returns the <i>n</i>'th Fibonacci number. <p>
      * 
      * @param  n Identifies the Fibonacci number to return; cannot be smaller than one.
      * @return The <i>n</i>'th Fibonacci number.
      * @throws IllegalArgumentException If {@code n} is smaller than one.
      */
    public BigInteger getNthFibonnaciNumber(int n) {
        if (n < 1) {
            throw new IllegalArgumentException("'n' must be one or greater");
        }
        // Fast to calculate, so we create a new sequence on each call, and
        // hence we do not have to synchronise...
        FibonacciSequence fs = new FibonacciSequence();
        for (int i = 0; i < n; i++) {
            fs.next();
        }
        return fs.current();
    }

    /**
      * Returns a random Fibonacci number between the first 
      * and the {@code n}'th number. <p>
      * 
      * @param  n The maximum number to return; cannot be smaller than one.
      * @return A random Fibonacci number between the first and the
      *         {@code n}'th number.
      * @throws IllegalArgumentException If {@code n} is smaller than one.
      */
    public BigInteger getRandomFibonnaciNumber(int n) {
        if (n < 1) {
            throw new IllegalArgumentException("'n' must be one or greater");
        }
        return this.getNthFibonnaciNumber(Math.max(1, (int)(Math.abs(this.getRandomNumber()) % n)));
    }
    
    /**
      * Returns the <i>n</i>'th prime number. <p>
      * 
      * @param  n Identifies the prime number to return; cannot be smaller than one.
      * @return The <i>n</i>'th prime number.
      * @throws IllegalArgumentException If {@code n} is smaller than one.
      */
    public synchronized int getNthPrimeNumber(int n) {
        if (n < 1) {
            throw new IllegalArgumentException("'n' must be one or greater");
        }
        // Lazy initialisation...
        if (this.primes == null) {
            this.primes = new ReversiblePrimeSequence(100);
        }
        
        int m = n;
        this.primes.reset();
        while (true) {
            // Use iterator to go through values...
            @ParticipantUsage(value = "SubsystemClass", type = SequenceIterator.class)
            Iterator<Integer> i = new SequenceIterator<Integer>(this.primes);
            while (i.hasNext()) {
                int prime = i.next();
                if (--m == 0) {
                    return prime; 
                }
            }
            // No match yet, so we "grow" primes - expensive!
            this.primes = new ReversiblePrimeSequence(this.primes.maximum() * 2);
            m = n;
        }
    }

    /**
      * Returns a random prime number between the first 
      * and the {@code n}'th prime number. <p>
      * 
      * @param  n The maximum number to return; cannot be smaller than one.
      * @return A random prime number between the first and the
      *         {@code n}'th prime number.
      * @throws IllegalArgumentException If {@code n} is smaller than one.
      */
    public int getRandomPrimeNumber(int n) {
        if (n < 1) {
            throw new IllegalArgumentException("'n' must be one or greater");
        }
        return this.getNthPrimeNumber(this.getRandomNumber(n));
    }
    
    /**
      * Returns the value of the Ackermann function, <i>A(m,n)</i>, for 
      * <i>m</i> and <i>n</i> values supplied as {@code m} and {@code n},
      * respectively. <p>
      * 
      * @param  m The <i>m</i> value for the Ackermann function; cannot be negative.
      * @param  n The <i>n</i> value for the Ackermann function; cannot be negative.
      * @return The <i>A(m,n)</i> value.
      * @throws IllegalArgumentException If {@code m} or {@code n} are illegal.
      */
    public int getAckermannNumber(int m, int n) {
        if (n < 0) {
            throw new IllegalArgumentException("'n' must be zero or greater");
        }
        AckermannSequence as = null;
        synchronized (this.ackermann) {
            if ((as = this.ackermann.get(m)) == null) {
                this.ackermann.put(m, as = new AckermannSequence(m));
            }
        }
        synchronized (as) {
            if (n > as.n()) {
                as.reset(); // 
            }
            for (int i = 0; i < n; i++) {
                as.next();
            }
            return as.current();
        }
    }

    /**
      * Returns a random number. 
      *  
      * @return A random number.
      */
    public long getRandomNumber() {
        return this.random.next();
    }
    
    /**
      * Return a pseudo random number between one and
      * {@code highBound}.
      * 
      * @param  highBound The high bound.
      * @return The pseudo random number.
      * @throws IllegalArgumentException If {@code highBound} is smaller than
      *         one.
      */
    public int getRandomNumber(int highBound) {
        return this.getRandomNumber(1, highBound);
    }
    
    /**
      * Return a pseudo random number between {@code lowBound} and
      * {@code highBound}.
      * 
      * @param  lowBound  The low bound.
      * @param  highBound The high bound.
      * @return The pseudo random number.
      * @throws IllegalArgumentException If {@code highBound} is smaller than
      *         {@code lowBound}.
      */
    public int getRandomNumber(int lowBound, int highBound) {
        if (lowBound > highBound) {
            throw new IllegalArgumentException("Illegal bounds");
        }
        // From zero to upper bound...
        RandomSequence rs = new RandomSequence(Math.max(highBound - lowBound, 1));
        // We can truncate because the higher bound is an integer
        // and random sequence uses longs...
        int random = rs.next().intValue();
        return random + lowBound;
    }

    /**
      * Returns a pseudo random boolean value.
      *   
      * @return True or false.
      */
    public boolean getRandomBoolean() {
        return (this.getRandomNumber() % 2 == 0);  
    }
    
    //
    // Lots of other complex math functions here...
    //
    
    // ..
    
    /**
      * Return the string representation of this facade.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        return ObjectPolicy.NAME.asString(this);
    }
}
