/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.factorymethod;

import dk.rode.thesis.command.Command;
import dk.rode.thesis.command.NextCommand;
import dk.rode.thesis.command.NullCommand;
import dk.rode.thesis.command.ResetCommand;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.model.Sequence.State;

/**
  * A <i>standard command creator</i> creates new
  * {@link Command commands} based on sequence {@link 
  * dk.rode.thesis.meta.model.Sequence.State states}
  * as tokens. <p>
  *
  * This creator can create the following commands based on the 
  * supplied internal sequence state: <ul>
  * 
  * <li> null: a {@link NullCommand} is returned. <p> 
  * 
  * <li> {@link dk.rode.thesis.meta.model.Sequence.State#NORMAL}: 
  *      creates a new {@link NextCommand} instance. <p>
  *       
  * <li> {@link dk.rode.thesis.meta.model.Sequence.State#RESET}: 
  *      creates a new {@link ResetCommand} instance. <p>
  *      
  * <li> Otherwise: a command created by {@link #createDefault(Sequence)}
  *      is returned, if not null; if null, a null command is returned. <p>
  *      
  * </ul>
  *  
  * A sequence command creator is stateless and thread-safe. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by created commands (and thus sequences).
  */
@Participant("ConcreteCreator")
public class SequenceCommandCreator<E> extends CommandCreator<E,State> {

    /**
      * No-arg constructor.
      */
    public SequenceCommandCreator() {
        super();
    }
    
    /**
      * 
      */
    @Override
    protected Command<E> create(Sequence<E> sequence, State state) throws Exception {
        if (state == null) {
            return null;
        }
        switch (state) {
            case NORMAL:
                return new NextCommand<E>(sequence);
            case RESET:
                return new ResetCommand<E>(sequence);
            default:
                return null;
        }
    }
        
}
