/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import static dk.rode.thesis.meta.util.Const.SERIAL_VERSION_UID;
import static dk.rode.thesis.meta.util.Util.add;

import java.util.ArrayList;
import java.util.List;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>break expression</i> allows the {@link Interpreter interpretation}
  * of a given expression syntax-tree to break at the point of the
  * break expression, and possibly continue the interpretation starting
  * from a specified <i>target expression</i>. Only a {@link TypedExpression}
  * instance can be the target. <p>
  * 
  * The target expression cannot be a break expression unless explicitly
  * cast to a {@code TypedExpression}, but a break expression can always 
  * be contained in the target expression. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * A break expression is a crude way to transfer evaluation control,
  * but quite effective for small languages though it depends on
  * an {@link Interpreter interpreter}. It utilises a
  * {@link BreakException break} exception, which is caught and
  * processed by the interpreter. The break exception stores the
  * target expression to transfer control to, if any. <p>  
  * 
  * However, exception types cannot be generic, so the expression 
  * to transfer control to must be declared using a wild-card in 
  * the {@code BreakException} type, i.e. {@code TypedExpression<?>}. 
  * Before the expression is interpreted, the interpreter will cast 
  * it into the proper type, {@code TypedExpression<E>}. This only 
  * works because we <b>only</b> allow {@link TypedExpression typed} 
  * expressions as the targets for break exceptions. Hence, we can test 
  * on the actual type at runtime, but the compiler does not know this when
  * we perform the cast, and we therefore have to suppress warnings 
  * explicitly.  
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of value the evaluation of this expression 
  *             produces.
  */
@Participant("NonTerminalExpression")
public class BreakExpression<E> 
     extends AbstractExpression<E>
  implements NonTerminalExpression<E>, TypedExpression<E> {

    /**
      * A <i>break exception</i> is thrown when an 
      * {@link BreakExpression break} expression is evaluated.  <p>
      * 
      * If the exception stores an expression, the 
      * {@link Interpreter interpreter} will transfer control to
      * that expression, or exit with a null result if the
      * exception does not store an expression. <p>
      *
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      */
    public final static class BreakException extends ExpressionException {

        /**
          * Serial version uid.
          */
        static final long serialVersionUID = SERIAL_VERSION_UID;

        /**
          * The target expression to be evaluated next. <p>
          * 
          * Can be null.
          */
        public final TypedExpression<?> expression;

        /**
          * No-arg constructor, which creates this break exception
          * to cause the interpreter to exit.
          */
        BreakException() {
            this(null);
        }
        
        /**
          * Constructor, which creates this break exception
          * to cause the interpreter to transfer control to 
          * {@code expression}, or exit if it is null.
          * 
          * @param expression The target expression; can be null.
          */
        BreakException(TypedExpression<?> expression) {
            super(expression);
            this.expression = expression;
        }
        
        /**
          * Return the string representation of this exception. 
          *
          * @return The string representation; never null.
          */
        public String toString() {
            StringBuilder sb = (StringBuilder)ObjectPolicy.NAME_OPEN.toString(this);
            if (this.expression != null) {
                sb.append("expression: ").append(this.expression);
            }
            return sb.append(']').toString();
        }
    }

    /**
      * The target expression to transfer control to, if any. <p>
      * 
      * Can be null.
      * 
      * @serial exclude
      */
    private final transient TypedExpression<E> expression; // not serialised!

    /**
      * No-arg constructor. <p>
      * 
      * When this expression is evaluated by an 
      * {@link Interpreter interpreter}, the interpreter
      * will exit will a null value.
      */
    public BreakExpression() {
        this((TypedExpression<E>)null); 
    }
    
    /**
      * Constructor. <p>
      * 
      * When this expression is evaluated by an 
      * {@link Interpreter interpreter}, the interpreter
      * will transfer control to {@code expression}, if 
      * not null, and return the result of that evaluation. <p>
      * 
      * If null, the interpreter will exit.
      * 
      * @param expression The target expression to use; can be null.
      */
    public BreakExpression(TypedExpression<E> expression) {
        this.expression = expression;
    }

    /**
      * Copy constructor. <p>
      * 
      * Note, that if a break expression is to be the target
      * expression, it must be cast to a {@code TypedExpression}
      * and set using the {@link #BreakExpression(TypedExpression)}
      * constructor. This constructor is a copy constructor that
      * simply copies {@code expression}.
      * 
      * @param  expression The expression to copy; cannot be null.
      * @throws NullPointerException If {@code expression} is null.
      */
    public BreakExpression(BreakExpression<E> expression) {
        this.expression = expression.expression == null 
                        ? null 
                        : expression.expression.copy();
    }

    /**
      * 
      */
    public BreakExpression<E> copy() {
        return new BreakExpression<E>(this);
    }
    
    /**
      * 
      */
    public List<Expression<?>> operands() {
        return add(new ArrayList<Expression<?>>(), this.expression);
    }

    /**
      * @return The type of the target expression, or null if no expression
      *         is stored.
      */
    public final Class<E> type() {
        return (this.expression != null) ? this.expression.type() : null; 
    }
    
    /**
      * @return {@code exit} <b>or</b> {@code break}.
      */
    public final String name() {
        return this.expression == null ? "exit" : "break";
    }
    
    /**
      * 
      */
    public E evaluate(Context context) throws ExpressionException {
        LogFactory.getLog(this).println("Evaluating: ", this.asSymbol(context.reset()));
        throw new BreakException(this.expression);
    }
    
    /**
      * Returns either {@code exit} or {@code break[x]}, where 
      * {@code x} is the symbol representation of the target expression 
      * used, if any.    
      */
    public String asSymbol(Context context) throws ExpressionException {
        if (this.expression == null) {
            return "exit";
        }
        StringBuilder sb = new StringBuilder();
        sb.append("break[");
        SymbolIdiom.asSymbol(context, this, this.expression, sb);
        return sb.append(']').toString();
    }
    
    /**
      * 
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        if (this.expression != null) {
            sb.append("expression: ").append(this.expression);
        }
        return sb.append(']').toString();
    }
    
}
