/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import java.util.List;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.prototype.StrictCopyable;

/**
  * An <i>expression</i> represents grammar rules in form
  * or {@link NonTerminalExpression non terminal expressions} 
  * and actual functionality in form of {@link TerminalExpression
  * terminal expressions} that will be enforced when the expression
  * is {@link #evaluate(Context) evaluated}. <p>
  * 
  * Expressions are considered immutable, but can be {@link #copy() copied}. 
  * Each unique expression type have a {@link #name() name} and each 
  * instance {@link #asSymbol(Context) symbolic} representation. Expressions
  * that cannot be constructed in a <i>single step</i> must be
  * {@link InitialisableExpression initialised} before use; once
  * initialised, they are considered immutable. <p>
  *  
  * Expressions may use other expressions as {@link #operands() operands}.
  * If a given expression <i>uses</i> or <i>contains</i> another expression 
  * can be tested using the {@link #contains(Expression)} method, based
  * on identity. An expression containing cyclic expression references 
  * can still be evaluated and generate a symbolic representation, but 
  * <b>cannot</b> be copied, queried for containment, or generate its 
  * {@code toString()} representation. <p>
  * 
  * {@code Expression} is the common super type for all expressions
  * that can be evaluated by an {@link Interpreter interpreter}. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * As expressions are considered immutable, all but a few internal 
  * attributes can be declared final. Simple immutable flags or values 
  * can therefore be exposed directly through a final public instance 
  * variable, such as the value of a {@link ConstantExpression#value constant} 
  * or the {@link CompareExpression#comparison comparison} type used. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    TypedExpression
  * @see    InitialisableExpression
  * 
  * @param  <E> The type of value the evaluation of this expression 
  *             produces.
  */
@Participant("AbstractExpression")
public interface Expression<E> extends StrictCopyable<Expression<E>> {

    /**
      * The <i>symbol idiom</i> ensures that cyclic expression
      * references will be represented correctly in 
      * {@link Expression#asSymbol(Context) symbolic}
      * representation starting from a given expression. 
      *
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      * 
      * @see    Context#reset()
      * @see    Context#touch(Expression)
      */
    public final static class SymbolIdiom { // has to be declared public in an interface!

        /**
          * Private off-limit constructor.
          */
        private SymbolIdiom() {
            throw new AssertionError(); // guard against reflection
        }
        
        /**
          * See {@link Expression#asSymbol(Context)}.
          *   
          * @param  context The context to use; cannot be null.
          * @param  parent  The parent expression; cannot be null.
          * @param  child   The child expression; cannot be null.
          * @return The symbolic representation; never null.
          * @throws NullPointerException If either argument is null.
          * @throws ExpressionException If the symbol cannot be generated.
          */
        public final static StringBuilder asSymbol(Context context, Expression<?> parent, Expression<?> child) throws ExpressionException {
            return asSymbol(context, parent, child, new StringBuilder());
        }

        /**
          * See {@link Expression#asSymbol(Context)}.
          *   
          * @param  context The context to use; cannot be null.
          * @param  parent  The parent expression; cannot be null.
          * @param  child   The child expression; cannot be null.
          * @param  sb      The string builder to append the
          *                 symbolic representation to; cannot be null.
          * @return {@code sb}; never null.
          * @throws NullPointerException If either argument is null.
          * @throws ExpressionException If the symbol cannot be generated.
          */
        public final static StringBuilder asSymbol(Context context, Expression<?> parent, Expression<?> child, StringBuilder sb) throws ExpressionException {
            if ((context.touch(child)) && (child.contains(parent))) {
                sb.append("{{").append(parent.name() + " -> " + child.name()).append("}}");
            } else {
                sb.append(child.asSymbol(context));
            }
            return sb;
        }
        
    }
    
    /**
      * Evaluates this expression and returns the result. <p>
      * 
      * There is <b>no</b> guarantee that the evaluation of
      * this expression will terminate!
      * 
      * @param  context The context to use; cannot be null.
      * @return The result of the evaluation; never null.
      * @throws NullPointerException If {@code context} is null. 
      * @throws ExpressionException If the evaluation fails.
      */
    public E evaluate(Context context) throws ExpressionException;
    
    /**
      * Copies this expression. <p>
      * 
      * Unlike {@link #asSymbol(Context)}, copying 
      * <b>cannot</b> handle cyclic expression references!
      * 
      * @return A copy of this expression; never null.
      */
    public Expression<E> copy();
 
    /**
      * Returns the expression <i>operands</i> used by this expression,
      * in order. <p>
      *  
      * Modifying the returned list will not affect this expression.
      *  
      * @return The operands; never null, but can be empty.
      */
    public List<Expression<?>> operands();
 
    /**
      * Returns true if this expression is or contains the 
      * expression supplied as {@code expression}, false if not. <p>
      * 
      * The test is performed using identity (==). <p>
      *  
      * Unlike {@link #asSymbol(Context)}, querying for containment
      * <b>cannot</b> handle cyclic expression references!
      * 
      * @param  expression The expression to test; cannot be null.
      * @return True if {@code expression} is this expression or
      *         contained within this expression, false if not.
      * @throws NullPointerException If {@code expression} is null.
      * @see    #asSymbol(Context)
      * @see    Context#touch(Expression)         
      */
    public boolean contains(Expression<?> expression);
    
    /**
      * Returns a short <i>symbolic</i> representation of this 
      * expression that describes it in a concise form. <p>
      * 
      * The caller of this method should <b>always</b> ensure
      * that {@code context} is reset properly before use
      * by calling {@link Context#reset()}. This ensures that
      * cyclic expression references are handled properly! <p>
      * 
      * General semantics: <ul><p>
      * 
      * <li> Expressions are separated by comma (,), and possibly grouped
      *      in parentheses. <p>
      * 
      * <li> Standard Java operators are used: {@code &&} and }||} 
      *      for logical <i>short-circuit</i> AND and OR, respectively;
      *      {@code &} and }|} for logical non <i>short-circuit</i> 
      *      AND and OR, respectively; {@code <}, {@code <=}, {@code ==}, 
      *      {@code !=}, {@code >=}, and {@code >} for comparison; {@code =} 
      *      for assignment; {@code !} for negation; and {@code x ? y : z} for 
      *      conditional expressions, where {@code x}, {@code y}, and {@code z} 
      *      are expressions. <p>
      * 
      * <li> Arguments to {@link TerminalExpression terminal} expressions
      *      are enclosed in square brackets, not parentheses, like 
      *      {@code x[y]}, where {@code x} is the terminal expression
      *      taking expression {@code y} as an argument. <p> 
      *      
      * <li> {@link TypedExpression typed} expressions may expose the type 
      *      like {@code Type{x}}, where {@code Type} is the simple 
      *      class name of the type, and {@code x} the symbolic 
      *      representation of the expression. <p>
      *       
      * <li> The symbolic representation of a cyclic reference is
      *      {@code {{from -> to}}}, where {@code from} and {@code to}
      *      are the {@link #name() names} of the expressions, respectively. <p>
      *      
      * </ul>
      * 
      * @param  context The context to use; never null.
      * @return The symbolic representation; never null.
      * @throws NullPointerException If {@code context} is null. 
      * @throws ExpressionException If the symbol cannot be generated.
      * @see    #contains(Expression)  
      */
    public String asSymbol(Context context) throws ExpressionException;

    /**
      * The stand-alone symbol name for this expression. 
      *   
      * @return The symbol name; never null.
      */
    public String name();
    
    /**
      * Returns the string representation of this expression. <p>
      * 
      * Unlike {@link #asSymbol(Context)}, the string generation
      * <b>cannot</b> handle cyclic expression references!
      *  
      * @return The string representation; never null.
      */
    public String toString();
}
