/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import static dk.rode.thesis.meta.util.Util.add;

import java.util.List;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.ReversibleSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.ObjectPolicy;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * A <i>reverse expression</i> will invoke 
  * {@link ReversibleSequence#reverse() reverse()}
  * on a given {@link ReversibleSequence reversible} sequence when 
  * evaluated, if so specified and <b>only</b> if possible. <p>
  * 
  * In order to reverse the sequence, the sequence must
  * not only be {@link ReversibleSequence#reversible() reversible},
  * but an expression can also be {@link #ReverseExpression(Sequence, Expression) 
  * supplied} to determine whether or not to reverse at evaluation time. 
  * Any expression of the type {@code Expression<Boolean>} can be used. 
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    CurrentExpression
  * @see    NextExpression
  * @see    SetExpression
  * @see    ResetExpression
  * 
  * @param  <E> The (super-) type of values delivered by the manipulated 
  *             {@link Sequence sequence}.
  */
@Participant("TerminalExpression")
public class ReverseExpression<E> extends SequenceExpression<E,Boolean> {
    
    /**
      * If this expression evaluates to true, the sequence will be
      * <i>tried</i> reversed, not if it evaluates to false.
      */
    private final Expression<Boolean> reverse;
    
    /**
      * Constructor. <p>
      * 
      * This expression will <i>try</i> and reverse {@code sequence}. <p>
      * 
      * The identity hash code of {@code sequence} is not included
      * in the {@link #asSymbol(Context) symbolic} representation of 
      * this expression.
      * 
      * @param  sequence The sequence to manipulate by this expression;
      *                  cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public ReverseExpression(Sequence<? extends E> sequence) {
        this(sequence, true);
    }
    
    /**
      * Constructor. <p>
      * 
      * This expression will <i>try</i> and reverse {@code sequence} if
      * {@code reverse} is true, not if false (no-op). <p>
      * 
      * The identity hash code of {@code sequence} is not included
      * in the {@link #asSymbol(Context) symbolic} representation of 
      * this expression.
      * 
      * @param  sequence The sequence to manipulate by this expression;
      *                  cannot be null.
      * @param  reverse  True to <i>try</i> and reverse {@code sequence},
      *                  false not to (no-op).
      * @throws NullPointerException If {@code sequence} is null.
      */
    public ReverseExpression(Sequence<? extends E> sequence, boolean reverse) {
        this(sequence, reverse, false);
    }
    
    /**
      * Constructor. <p>
      * 
      * This expression will <i>try</i> and reverse {@code sequence} if
      * {@code reverse} is true, not if false (no-op). <p>
      * 
      * The identity hash code of {@code sequence} is included
      * in the {@link #asSymbol(Context) symbolic} representation of 
      * this expression if {@code includeHashCode} is true.
      * 
      * @param  sequence        The sequence to manipulate by this expression;
      *                         cannot be null.
      * @param  reverse         True to <i>try</i> and reverse {@code sequence},
      *                         false not to (no-op).
      * @param  includeHashCode True to include the identity hash code
      *                         of {@code sequence} in the symbolic
      *                         representation, false not to.                 
      * @throws NullPointerException If {@code sequence} is null.
      */
    public ReverseExpression(Sequence<? extends E> sequence, boolean reverse, boolean includeHashCode) {
        this(sequence, new ConstantExpression<Boolean>(Boolean.class, reverse), includeHashCode);
    }
    
    /**
      * Constructor. <p>
      * 
      * This expression will <i>try</i> and reverse {@code sequence} if
      * {@code reverse} evaluates to true when this expression is
      * evaluated. <p>
      * 
      * The identity hash code of {@code sequence} is not included
      * in the {@link #asSymbol(Context) symbolic} representation of 
      * this expression.
      * 
      * @param  sequence The sequence to manipulate by this expression;
      *                  cannot be null.
      * @param  reverse  The expression that will determine if {@code sequence}
      *                  is <i>tried</i> reversed when this expression is evaluated.
      * @throws NullPointerException If either argument is null.
      */
    public ReverseExpression(Sequence<? extends E> sequence, Expression<Boolean> reverse) {
        this(sequence, reverse, false);    
    }
    
    /**
      * Constructor. <p>
      * 
      * This expression will <i>try</i> and reverse {@code sequence} if
      * {@code reverse} evaluates to true when this expression is
      * evaluated. <p>
      * 
      * The identity hash code of {@code sequence} is included
      * in the {@link #asSymbol(Context) symbolic} representation of 
      * this expression if {@code includeHashCode} is true.
      * 
      * @param  sequence        The sequence to manipulate by this expression;
      *                         cannot be null.
      * @param  reverse         The expression that will determine if {@code sequence}
      *                         is <i>tried</i> reversed when this expression is evaluated.
      * @param  includeHashCode True to include the identity hash code
      *                         of {@code sequence} in the symbolic
      *                         representation, false not to.                 
      * @throws NullPointerException If {@code sequence} or {@code reverse} are null.
      */
    public ReverseExpression(Sequence<? extends E> sequence, Expression<Boolean> reverse, boolean includeHashCode) {
        super(sequence, includeHashCode);
        if ((this.reverse = reverse) == null) {
            throw new NullPointerException("Reverse expression cannot be null");
        }
    }
    
    /**
      * Copy constructor. <p>
      * 
      * The actual {@link Sequence sequence} manipulated by
      * {@code expression} is <b>not</b> copied, but used
      * as is.
      * 
      * @param  expression The expression to copy; cannot be null.
      * @throws NullPointerException If {@code expression} is null.
      */
    public ReverseExpression(ReverseExpression<E> expression) {
        super(expression);
        this.reverse = expression.reverse.copy();
    }
 
    /**
      *
      */
    public ReverseExpression<E> copy() {
        return new ReverseExpression<E>(this);
    }

    /**
      * 
      */
    @Override
    public List<Expression<?>> operands() {
        return add(super.operands(), this.reverse);
    }
    
    /**
      * @return {@code reverse}
      */
    public final String name() {
        return "reverse";
    }
    
    /**
      * 
      * @return True if the manipulated sequence could be
      *         reversed, false if not.
      * @see    ReversibleSequence#reverse()          
      */
    public Boolean evaluate(Context context) throws ExpressionException {
        String s = this.asSymbol(context.reset());
        Log out = LogFactory.getLog(this).println("Evaluating: ", s);
        boolean reversed = false,
                reverse = this.reverse.evaluate(context);
        if ((reverse) && (this.sequence instanceof ReversibleSequence)) {
            ReversibleSequence<?> rs = (ReversibleSequence<?>)this.sequence;
            if (rs.reversible()) {
                rs.reverse();
                reversed = true;
            }
        }
        out.println(s, " -> ", reversed);
        return reversed; 
    }
    
    /**
      * Returns {@code reverse[x]}, where {@code x} is the symbolic
      * representation of the expression used to determine if the
      * sequence should be reversed, or {@code reverse#123[x]}, 
      * where {@code 123} is the identity hash code of the 
      * manipulated sequence.  
      */
    @Override
    public String asSymbol(Context context) throws ExpressionException {
        StringBuilder sb = new StringBuilder();
        sb.append(super.asSymbol(context)).append('[');
        SymbolIdiom.asSymbol(context, this, this.reverse, sb);
        return sb.append(']').toString();
    }
    
    /**
      * 
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; sequence: ").append(this.sequence.toString(SequencePolicy.IDENTITY))
                 .append("; reverse: ").append(this.reverse).append(']').toString();
    }
    
}
