/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import static dk.rode.thesis.meta.util.Util.add;

import java.util.List;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.model.Sequence.State;
import dk.rode.thesis.strategy.ObjectPolicy;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * A <i>set expression</i> will <i>fast-forward</i> the value 
  * of a given <b>bounded</b> {@link Sequence sequence} to match 
  * a specific value, if possible. <p>
  * 
  * The value to fast-forward to can be a 
  * {@link #SetExpression(Sequence, Class, Comparable) constant} 
  * value, or supplied as an {@link #SetExpression(Sequence, Expression) 
  * expression} of type {@code Expression<Comparable>}. The 
  * evaluation will fail if no such value can be found at evaluation 
  * time! <p>  
  *
  * <b>Implementation notes</b>:<br>
  * The value is set by invoking {@link Sequence#next()} until a 
  * matching value is found, or until all sequence values have
  * been traversed and a match cannot be found, in which case
  * an exception is thrown. The worst-case running time is thus
  * proportional to the number of values in the sequence, <i>n</i>,
  * i.e. <i>O(n)</i>.   
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    CurrentExpression
  * @see    NextExpression
  * @see    ResetExpression
  * @see    ReverseExpression
  * 
  * @param  <E> The type of value the evaluation of this expression 
  *             produces. The value corresponds to the (super-) type of 
  *             values delivered by the manipulated {@link Sequence sequence}.
  */
@Participant("TerminalExpression")
public class SetExpression<E extends Comparable<? super E>> extends SequenceExpression<E,E> {

    /**
      * The comparable value to set for the sequence. <p>
      * 
      * Never null.
      */
    private final Expression<E> value;
    
    /**
      * Constructor. <p>
      * 
      * The value to <i>fast-forward</i> to is supplied as the 
      * constant {@code value}. <p>
      * 
      * The identity hash code of {@code sequence} is not included
      * in the {@link #asSymbol(Context) symbolic} representation of 
      * this expression.
      * 
      * @param  sequence The sequence manipulated by this expression;
      *                  cannot be null.
      * @param  type     The type of {@code value}; cannot be null.                  
      * @param  value    The value to <i>fast-forward</i> to; 
      *                  cannot be null.                  
      * @throws NullPointerException If either argument is null.
      * @throws IllegalArgumentException If {@code sequence} is not bounded.
      */
    public SetExpression(Sequence<? extends E> sequence, Class<E> type, E value) {
        this(sequence, type, value, false);
    }
    
    /**
      * Constructor. <p>
      * 
      * The value to <i>fast-forward</i> to is supplied as the 
      * constant {@code value}. <p>
      * 
      * The identity hash code of {@code sequence} is included
      * in the {@link #asSymbol(Context) symbolic} representation of 
      * this expression if {@code includeHashCode} is true.
      * 
      * @param  sequence        The sequence manipulated by this expression;
      *                         cannot be null.
      * @param  type            The type of {@code value}; cannot be null.                  
      * @param  value           The value to <i>fast-forward</i> to; 
      *                         cannot be null.                  
      * @param  includeHashCode True to include the identity hash code
      *                         of {@code sequence} in the symbolic
      *                         representation, false not to.                 
      * @throws NullPointerException If {@code sequence}, {@code type}, or {@code value}
      *         are null.
      * @throws IllegalArgumentException If {@code sequence} is not bounded.
      */
    public SetExpression(Sequence<? extends E> sequence, Class<E> type, E value, boolean includeHashCode) {
        this(sequence, new ConstantExpression<E>(type, value), includeHashCode);
    }
    
    /**
      * Constructor. <p> 
      * 
      * The identity hash code of {@code sequence} is not included
      * in the {@link #asSymbol(Context) symbolic} representation of 
      * this expression.
      * 
      * @param  sequence The sequence manipulated by this expression;
      *                  cannot be null.
      * @param  value    The expression that will determine the value to
      *                  <i>fast-forward</i> to for {@code sequence};
      *                  cannot be null.
      * @throws NullPointerException If either argument is null.
      * @throws IllegalArgumentException If {@code sequence} is not bounded.
      */
    public SetExpression(Sequence<? extends E> sequence, Expression<E> value) {
        this(sequence, value, false);
    }
    
    /**
      * Constructor. <p> 
      * 
      * The identity hash code of {@code sequence} is included
      * in the {@link #asSymbol(Context) symbolic} representation of 
      * this expression if {@code includeHashCode} is true.
      * 
      * @param  sequence        The sequence manipulated by this expression;
      *                         cannot be null.
      * @param  value           The expression that will determine the value to
      *                         <i>fast-forward</i> to for {@code sequence};
      *                         cannot be null.
      * @param  includeHashCode True to include the identity hash code
      *                         of {@code sequence} in the symbolic
      *                         representation, false not to.                 
      * @throws NullPointerException If {@code sequence} or {@code value}
      *         are null.
      * @throws IllegalArgumentException If {@code sequence} is not bounded.
      */
    public SetExpression(Sequence<? extends E> sequence, Expression<E> value, boolean includeHashCode) {
        super(sequence, includeHashCode);
        if (!this.sequence.bounded()) {
            throw new IllegalArgumentException("Sequence must be bounded");
        } else if ((this.value = value) == null) {
            throw new NullPointerException("Value expression cannot be null");
        }
    }
    
    /**
      * Copy constructor. <p>
      * 
      * The actual {@link Sequence sequence} manipulated by
      * {@code expression} is <b>not</b> copied, but used
      * as is.
      * 
      * @param  expression The expression to copy; cannot be null.
      * @throws NullPointerException If {@code expression} is null.
      */
    public SetExpression(SetExpression<E> expression) {
        super(expression);
        this.value = expression.value.copy();
    }
 
    /**
      *
      */
    public SetExpression<E> copy() {
        return new SetExpression<E>(this);
    }
    
    /**
      *
      */
    @Override
    public List<Expression<?>> operands() {
        return add(super.operands(), this.value);
    }
    
    /**
      * @return {@code set}
      */
    public final String name() {
        return "set";
    }
    
    /**
      * @throws ExpressionException If the value to <i>fast-forward</i>
      *         to cannot be matched to a sequence value.
      */
    public E evaluate(Context context) throws ExpressionException {
        // This will modify the sequence even if a match is not
        // found! But, since we throw an exception, it will not
        // matter as the evaluation will be aborted here after...
        
        String s = this.asSymbol(context.reset());
        Log out = LogFactory.getLog(this).println("Evaluating: ", s);
        // Fetch value to set...
        E value = this.value.evaluate(context);
        // Have we passed the value? A reversed sequence will still work,
        // though slower...
        if (value.compareTo(this.sequence.current()) < 0) {
            this.sequence.reset();
        }
        int restart = 0;
        while (value.compareTo(this.sequence.current()) != 0) {
            this.sequence.next();
            if (this.sequence.state() == State.RESTART) {
                if (++restart == 2) {
                    out.error("Cannot find sequence value: ", this.sequence, " -> ", value);
                    throw new ExpressionException("Cannot set value: " + value);
                }
            }
        }
        out.println(s, " -> ", value);
        return value;
    }

    /**
      * Returns {@code set[x]}, where {@code x} is the symbolic
      * representation of the expression that will determine the 
      * value, or {@code set#123[x]}, where {@code 123} is the
      * identity hash code of the manipulated sequence.    
      */
    @Override
    public String asSymbol(Context context) throws ExpressionException {
        StringBuilder sb = new StringBuilder();
        sb.append(super.asSymbol(context)).append('[');
        SymbolIdiom.asSymbol(context, this, this.value, sb);
        return sb.append(']').toString();
    }
    
    /**
      * 
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; sequence: ").append(this.sequence.toString(SequencePolicy.IDENTITY))
                 .append("; value: ").append(this.value).append(']').toString();
    }
    
}
