/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.reflect;

import static dk.rode.thesis.meta.util.Const.CONSTRUCTOR_NAME;
import static dk.rode.thesis.meta.util.Const.STATIC_INITIALISER_NAME;
import static dk.rode.thesis.meta.util.Util.copy;
import static dk.rode.thesis.meta.util.Throwables.toRuntimeException;

import java.lang.reflect.Member;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
  * A <i>caller</i> represents one or more {@link java.lang.reflect.Member
  * java.lang.reflect.Member} instances representing a specific <i>calling 
  * context</i>. <p>
  * 
  * The calling context can be either a static initialisation block,
  * an instance initialisation block, a constructor, or a method. The
  * caller thus represents a corresponding {@code Member} instance,
  * if possible. <p>
  * 
  * However, the member may <b>not</b> be able to identified uniquely 
  * based on the stack trace. It will not be possible to uniquely resolve a 
  * member in case the calling context represents an overloaded method; a 
  * constructor and/or instance initialisation block declared in a class 
  * that declares multiple constructors and/or initialisation blocks; or 
  * a static initialisation block declared in a class that declares several 
  * static initialisation blocks. <p>
  * 
  * The calling context is stored in the {@link #callerMember} attribute,
  * while the member representing the immediate calling context is stored
  * in the {@link #calleeMember} attribute. The attributes are never null,
  * but may not be able to identify the unique member, as stated. They can,
  * however, still be used to test on the different types of members, and
  * list the possible {@link MemberContext#getCandidates() candidates}
  * as well. <p>
  * 
  * Callers are immutable. <p>   
  * 
  * <b>Example</b>:
  * <pre>
  *   ..
  *   public class Goo {
  *     public void goo {
  *       // Calling context at depth 2...
  *       new Foo().foo();
  *     }
  *   }
  *   ..
  *   public class Foo {
  *      public void foo {
  *         // Calling context at depth 1...
  *         new Bar().bar();
  *      }
  *   }
  *   ..
  *   public class Bar {
  *     public void bar {
  *       // Immediate calling context is the callee, i.e. the
  *       // member at depth 0 (method bar)...
  *       Caller caller = new {@link #Caller()};  // = 1             
  *       System.out.println(caller.{@link #callerMember}); // prints "foo" 
  *       System.out.println(caller.{@link #calleeMember}); // prints "bar"
  *       
  *       caller = new {@link #Caller(int) Caller}(1);             
  *       System.out.println(caller.{@link #callerMember}); // prints "foo" 
  *       System.out.println(caller.{@link #calleeMember}); // prints "bar"
  *        
  *       caller = new {@link #Caller(int) Caller}(2);             
  *       System.out.println(caller.{@link #callerMember}); // prints "goo" 
  *       System.out.println(caller.{@link #calleeMember}); // prints "bar" 
  *     }
  *   }
  *   ..
  * </pre>
  * 
  * Notice that the no-arg constructor does <b>not</b> identify the 
  * immediate calling context as the {@link #callerMember}, but the 
  * calling context at stack depth 1. The immediate calling context 
  * is always available via the {@link #calleeMember} method. <p>
  * 
  * An {@link IdentifiedStackTraceElement#offset offset} can be supplied
  * at {@link #Caller(int) construction} time to specify a stack 
  * depth to identify the calling context. The offset is relative to
  * the immediate calling context. <p> 
  *  
  * Stack traces are local to the thread in question.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class Caller extends CallerClass {

    /**
      * A <i>caller type</i> identifies possible <i>caller contexts</i>
      * a given {@link Caller caller} represents.
      *  
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      */
    public enum CallerType {
        /**
          * The calling context represented by the caller is a 
          * static initialisation block.  
          */
        INITIALISATION, 
        /**
          * The calling context represented by the caller is a 
          * constructor or an instance initialisation block.  
          */
        CONSTRUCTION, 
        /**
          * The calling context represented by the caller is a 
          * method, static or otherwise.  
          */
        INVOCATION;
    }
    
    /**
      * A <i>member context</i> represents a member in a given 
      * <i>calling context</i>, which may or may not be uniquely
      * {@link #resolved}. <p>
      * 
      * The member represents a static initialiser block, an instance 
      * initialiser block, a constructor, or a method. The type is 
      * determined via the {@link #type} attribute. <p>
      *
      * If the member is uniquely resolved, the resolved member is 
      * stored in the {@link #member} attribute, and the {@link #resolved}
      * attribute is true. If not uniquely resolved, {@code resolved}
      * is false, and {@code member} contains the <b>first</b> matching
      * {@link #getCandidates() candidate}. <p>
      * 
      * It will not be possible to uniquely resolve a member in case 
      * the calling context represents an overloaded method; a constructor 
      * and/or instance initialisation block declared in a class that declares 
      * multiple constructors and/or initialisation blocks; or a static 
      * initialisation block declared in a class that declares several 
      * static initialisation blocks. Furthermore, there is no {@code Member} 
      * instance corresponding to an initialisation block, static or otherwise. 
      * For instance initialisation blocks, the {@code Constructor} class
      * is used. For static initialisation blocks, {@code member} is null 
      * and the candidate array is empty, and {@code type} is set to 
      * {@link CallerType#INITIALISATION}. <p>
      *    
      * The line number of the member in question is unique and 
      * is stored in the {@link #lineNumber} attribute. <p>
      * 
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      */
    public final static class MemberContext {
        
        /**
          * The stack trace element element identifying the 
          * member for this member context. <p>
          * 
          * Never null.
          */
        private final StackTraceElement element;
        /**
          * The identified member type of this member context. <p>
          * 
          * Never null.
          */
        public final CallerType type;
        /**
          * The member represented by this member context. <p>
          * 
          * If {@link #resolved} is true, the member is uniquely
          * resolved, not if false. <p>
          * 
          * If {@code resolved} is false, the possible 
          * candidates can be retrieved with {@link #getCandidates()},
          * and the <b>first</b> member from the candidate array
          * is stored in this attribute. <p>
          * 
          * Will be null in case this member context represents
          * a static initialisation block, but a {@code Constructor}
          * instance in case this member context represents 
          * an instance initialisation block.
          */
        public final Member member;
        /**
          * The possible member candidates for the member represented
          * by this member context. <p>
          * 
          * If {@link #resolved} is true, the single element contained
          * is {@link #member}. <p>
          * 
          * If this member context represents a static initialisation
          * block, the array is empty. <p>
          * 
          * The order of the candidates is their declaration order.
          */
        protected final Member[] candidates;
        /**
          * True if {@link #member} has been uniquely resolved
          * for this member context, false if not. 
          */
        public final boolean resolved; 
        /**
          * The line number of the stack trace element identifying 
          * the member used by this member context.  
          */
        public final int lineNumber;
        
        /**
          * Constructor, which creates this member context to
          * represent the member identified by the stack
          * trace element supplied as {@code element}.
          * 
          * @param  element The stack trace element identifying the
          *                member; cannot be null.
          * @throws NullPointerException If {@code element} is null.                
          */
        MemberContext(StackTraceElement element) {
            this.candidates = getMember(this.element = element);
            this.resolved = (this.candidates.length == 1);
            if (this.candidates.length > 0) {
                this.member = this.candidates[0];
                this.type = (this.member instanceof Method) 
                          ? CallerType.INVOCATION 
                          : CallerType.CONSTRUCTION;
            } else {
                this.member = null; // static
                this.type = CallerType.INITIALISATION;
            }
            this.lineNumber = element.getLineNumber();
        }
        
        /**
          * Returns the possible candidates that match the member 
          * this member context tries to identify. <p>
          * 
          * If {@link #resolved} is true, the returned array will
          * only contain {@link #member}. <p>
          * 
          * If this member context represents a static initialisation
          * block, the returned array will be empty. If it represents
          * an instance initialisation block, the returned array
          * contains all declared constructors. <p>
          * 
          * The order of the candidates is their declaration order. <p>
          * 
          * Modifying the returned array will not affect this member 
          * context.
          * 
          * @return The candidates; never null, but but can be empty if 
          *         this member context represents a static initialisation 
          *         block.
          */
        public Member[] getCandidates() {
            return copy(this.candidates, new Member[this.candidates.length]);
        }
        
        /**
          * Returns the hash code of this member context.
          * 
          * @return The hash code.
          */
        public int hashCode() {
            return this.element.hashCode();
        }
        
        /**
          * Returns true if {@code object} is an instance
          * of this class identifying a member based on an
          * equal stack trace element, false if not.
          */
        public boolean equals(Object object) {
            if (!(object instanceof MemberContext)) {
                return false;
            }
            MemberContext mc = (MemberContext)object;
            return this.element.equals(mc);
        }

        /**
          * Returns the string representation of this member context. <p>
          * 
          * The representation corresponds to that of the stack 
          * trace element used to identify the actual member.
          * 
          * @return The string representation; never null.
          */
        public String toString() {
            return this.element.toString();
        }
    }
    
    /**
      * The <i>caller member context</i> represents the member of the 
      * <i>calling context</i>, relative to the 
      * {@link IdentifiedStackTraceElement#offset offset} supplied 
      * at construction time. <p>
      * 
      * The member is the "other member" to test when methods 
      * like {@link #isConstructor()} is invoked,
      * where {@link #calleeMember} represents "this class". <p>
      * 
      * Never null.
      * 
      * @see CallerClass#callerClass
      */
    public final MemberContext callerMember;
    /**
      * The <i>callee member context</i> represents the member that created
      * this caller instance, i.e. the member representing the the <i>immediate 
      * calling context</i>. <p>
      * 
      * Never null.
      * 
      * @see CallerClass#calleeClass
      */
    public final MemberContext calleeMember;
    
    /**
      * No-arg constructor. <p>
      * 
      * The caller {@link #callerMember member} is the member of the calling 
      * context at stack depth 1 from the {@link #calleeMember member} of 
      * the immediate calling context, i.e. the calling context of the 
      * immediate calling context. <p> 
      * 
      * <b>Example</b>:
      * <pre>
      *   ..
      *   public void foo {
      *     this.bar();
      *   }
      * 
      *   public void bar {
      *     // Represents foo, *not* bar
      *     Caller caller = new Caller(); 
      *   }
      *   ..
      * </pre>
      */  
    public Caller() {
        // +1 for this class and +1 for this constructor...
        this(1, 1);  
    }
  
    /**
      * Constructor. <p>
      * 
      * The caller {@link #callerMember member} is the member of the calling 
      * context at stack depth {@code offset} from the {@link #calleeMember 
      * member} of the immediate calling context. <p> 
      *  
      * Invoking this constructor with a value of zero corresponds to invoking 
      * the no-arg constructor.
      *   
      * @param  offset The stack offset; must be within bounds.
      * @throws IndexOutOfBoundsException If {@code offset} is illegal, i.e., does not
      *         represent a valid stack trace element. 
      */
    public Caller(int offset) {
        // +1 for this class and +1 for this constructor...
        this(1, offset + 1);
    }
    
    /**
      * Constructor. <p>
      * 
      * The caller {@link #callerMember member} is the member of the calling 
      * context at stack depth {@code offset} from the {@link #calleeMember 
      * member} of the immediate calling context. <p> 
      *  
      * Invoking this constructor with zero values corresponds to invoking 
      * the no-arg constructor.
      *   
      * @param  depth  The increase in stack depth added by this class or
      *                by sub-classes. The depth of this class is zero,
      *                the depth of an immediate sub-class is one, and
      *                so forth.
      * @param  offset The stack offset; must be within bounds.
      * @throws IndexOutOfBoundsException If {@code offset} is illegal, i.e., does not
      *         represent a valid stack trace element. 
      */
    protected Caller(int depth, int offset) {
        // +1 for this class and +1 for this constructor...
        super(depth, offset + 1);
        // Fetch specified caller member...
        this.callerMember = new MemberContext(this.element);
        // Fetch the member of the calling context that
        // created this caller class. We add 3 to the depth 
        // because we know that the super-class adds three
        // stack frames!
        this.calleeMember = new MemberContext(this.stack[depth + 3]);
    }
    
    /**
      * Constructor, which creates this caller using an
      * existing stack trace supplied from {@code caller}, 
      * but using a perhaps different offset supplied as 
      * {@code offset}. <p>
      * 
      * @param  caller The caller to supply the stack trace; 
      *                cannot be null.
      * @param  offset The stack offset; must be within bounds.
      * @throws NullPointerException If {@code caller} is null.
      * @throws IndexOutOfBoundsException If {@code offset} is illegal, i.e., 
      *         does not represent a valid stack trace element. 
      */
    protected Caller(Caller caller, int offset) {
        super(caller, offset);
        // Fetch specified caller member...
        this.callerMember = new MemberContext(this.element);
        // The member of the calling context that
        // created this caller class have not changed,
        // so we simply reuse it...
        this.calleeMember = caller.calleeMember;
    }
    
    /**
      * Copy constructor. 
      * 
      * @param  caller The caller to copy; cannot be null.
      * @throws NullPointerException If {@code caller} is null.
      */
    public Caller(Caller caller) {
        this(caller, caller.offset);
    }
    
    /**
      * Returns the members associated with the stack trace element
      * supplied as {@code element}. The members will always be
      * constructors or methods. <p>
      *   
      * The order of the members is their declaration order in the
      * class represented by {@code element}, and only members
      * from that class are considered, i.e. no inherited members.
      * 
      * @param  element The stack trace element; cannot be null.
      * @return The matching members; never null.
      * @throws NullPointerException If {@code element} is null.
      */
    protected final static Member[] getMember(StackTraceElement element) {
        try {
            // Get class...
            Class<?> clazz = getClass(element);
            // Method name...
            String name = element.getMethodName();
            
            // Static initialisation block... 
            if (name.indexOf(STATIC_INITIALISER_NAME) != -1) {
                return new Member[0];
                
            // Constructor or instance initialisation block...
            } else if (name.indexOf(CONSTRUCTOR_NAME) != -1) { 
                return clazz.getDeclaredConstructors();
                
            // Method...
            } else {
                List<Member> list = new ArrayList<Member>(3);
                for (Method m : clazz.getDeclaredMethods()) {
                    if (m.getName().equals(name)) {
                        list.add(m);
                    }
                }
                Member[] members = new Member[list.size()];
                list.toArray(members);
                return members;
            }
                        
        } catch (Throwable throwable) {
            // Should never happen...
            throw toRuntimeException(throwable);
        }
    }
    
    /**
      * Returns true if the {@link #callerMember member} of the calling context 
      * is a static initialiser block, false if not. <p> 
      *  
      * <b>Example</b>: 
      * <pre>
      *   class Foo {
      *
      *     static {
      *       bar();
      *     } 
      * 
      *     public void bar {
      *       Caller caller = new Caller();
      *       boolean isStatic = caller.isStatic(); // True...
      *       ..
      *     }
      * 
      *     ..
      *   }
      * </pre>
      * 
      * To test if a given static method is the caller, use the 
      * {@link #isMethod()} instead, and then test on method
      * modifiers.
      *  
      * @return True if the member of the calling context is a static 
      *         initialiser block, false if not. <p> 
      * @see    #isConstructor()
      * @see    #isMethod()         
      */
    public final boolean isStatic() {
        return (this.callerMember.type == CallerType.INITIALISATION);
    }
  
    /**
      * Returns true if the {@link #callerMember member} of the calling context 
      * is a constructor or instance initialisation block, false if not. <p> 
      *  
      * <b>Example</b>: 
      * <pre>
      *   class Foo {
      * 
      *     {
      *       bar();
      *     }
      *     
      *     public Foo {
      *       bar();
      *     }
      * 
      *     public static void bar {
      *       Caller caller = new Caller();
      *       boolean isConstructor = caller.isConstructor(); // True, True...
      *       ..
      *     }
      * 
      *     ..
      *   }
      * </pre>
      *  
      * @return True if the member of the calling context is a constructor or 
      *         instance initialisation block, false if not. <p> 
      * @see    #isStatic()
      * @see    #isMethod()         
      */
    public final boolean isConstructor() {
        return (this.callerMember.type == CallerType.CONSTRUCTION);
    }
  
    /**
      * Returns true if the {@link #callerMember member} of the calling context 
      * is a method, false if not. <p> 
      *  
      * <b>Example</b>: 
      * <pre>
      *   ..
      *   public void foo {
      *     this.bar();
      *   }
      * 
      *   public void bar {
      *     Caller caller = new Caller();
      *     boolean isMethod = caller.isMethod(); // True: foo is a method...
      *     ..
      *   }
      *   ..
      * </pre>
      * 
      * @return True if the member of the calling context is a method, false 
      *         if not. 
      * @see    #isStatic()
      * @see    #isConstructor()         
      */
    public final boolean isMethod() {
        return (this.callerMember.type == CallerType.INVOCATION);
    }
    
    /**
      * Returns true if just one of the declared methods and constructors 
      * fetched from an object in {@code objects} is contained in the 
      * call-stack of this caller, false if not. <p>
      * 
      * {@code objects} may contain {@code Class} objects as well.
      *  
      * @param  objects The objects delivering the methods and constructors to 
      *                 test; cannot contain null entries.
      * @return True if just one of the methods or constructors fetched from 
      *         an object in {@code objects} is contained in the call-stack, 
      *         false if not.
      * @throws NullPointerException If {@code objects} contains null entries.
      * @see    #isMemberInCallStack(Member...)         
      */
    public boolean isMemberInCallStack(Object... objects) {
        for (Object o : objects) {
            Class<?> c = o instanceof Class ? (Class<?>)o : o.getClass();
            // Methods...
            if (this.isMemberInCallStack(c.getDeclaredMethods())) {
                return true;
            // Constructors...
            } else if (this.isMemberInCallStack(c.getDeclaredConstructors())) {
                return true;
            } 
        }
        return false;
    }
  
    /**
      * Returns true if just one of the members supplied in {@code members} 
      * is contained in the call-stack of this caller, false if not. <p>
      * 
      * Only {@code Method} or {@code Constructor} members can ever 
      * result in a match, while all other member types will cause
      * false to be returned. 
      * 
      * @param  members The members to test; cannot contain null entries.
      * @return True if just one of the members in {@code members} is 
      *         contained in the call-stack, false if not.
      * @throws NullPointerException If {@code members} contains null entries.
      * @see    #isMemberInCallStack(Object...)         
      */
    public boolean isMemberInCallStack(Member... members) {
        for (Member member : members) {
            for (int j = this.stack.length - 1; j >= this.offset; j--) {
                MemberContext mc = new MemberContext(this.stack[j]);
                if (mc.resolved) {
                    if (mc.member.equals(member)) {
                        return true;
                    }
                } else {
                    for (Member m : mc.candidates) {
                        if (m.equals(member)) {
                            return true;       
                        }
                    }
                }
            }
        }
        return false;
    }
    
    /**
      * 
      */
    @Override
    public Caller getSubClassCaller() {
        return (Caller)super.getSubClassCaller();
    }

    /**
      * 
      */
    @Override
    public Caller getNonSubClassCaller() {
        return (Caller)super.getNonSubClassCaller();
    }
    
    /**
      * 
      */
    @Override
    public Caller getAssignableCaller() {
        return (Caller)super.getAssignableCaller();
    }
    
    /**
      * 
      */
    @Override
    public Caller getNonAssignableCaller() {
        return (Caller)super.getNonAssignableCaller();
    }

    /**
      * 
      */
    @Override
    public Caller getClassCaller() {
        return (Caller)super.getClassCaller();
    }
    
    /**
      * 
      */
    @Override
    public Caller getNonClassCaller() {
        return (Caller)super.getNonClassCaller();
    }
    
    /**
      * Returns the <b>first</b> caller of this caller
      * found in the call-stack matching the caller type
      * supplied as {@code type}, if any. <p>
      *
      * @param  type The caller type; cannot be null.
      * @return The first matching caller, or null.
      * @throws NullPointerException If {@code type} is null.
      */
    public Caller getCaller(CallerType type) {
        for (Iterator<Caller> i = this.getCallers(); i.hasNext(); ) {
            Caller caller = i.next();
            if (caller.callerMember.type == type) {
                return caller;
            }
        }
        return null;
    }

    /**
      * 
      */
    @Override
    public Caller getCaller() {
        if (this.stack.length > this.offset) {
            return new Caller(this, this.offset + 1);
        }
        return null;
    }
    
    /**
      * 
      */
    @Override
    public Iterator<Caller> getCallers() {
        return new CallerIterator<Caller>(this) {
            protected void setNext() {
                this.caller = this.caller.getCaller();
            }
        };
    }
           
}
