/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.reflect;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import dk.rode.thesis.meta.util.Primitive;

/**
  * The <i>executor processor</i> provides methods to retrieve methods
  * and/or constructors annotated with the {@link Executor} annotation.   
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    Executor
  */
public class ExecutorProcessor {

    /**
      * Private off-limit constructor.
      */
    private ExecutorProcessor() {
        throw new AssertionError();
    }
    
    /**
      * Returns all declared methods annotated with the {@link Executor} annotation 
      * in the class supplied as {@code clazz}, where the formal parameter types
      * match {@code parameterTypes} ignoring primitives, and where the 
      * {@link Executor#value() context} is assignable from {@code context}. 
      *  
      * @param  clazz          The class; cannot be null.
      * @param  context        The context; cannot be null.
      * @param  parameterTypes The formal parameter types for the methods; can be empty.
      * @return A list containing the proper annotated methods in {@code clazz}, if
      *         any; never null, but may be empty.
      * @throws NullPointerException If {@code clazz} or {@code context} are null.         
      */
    public final static List<Method> getMethodExecutors(Class<?> clazz, Class<?> context, Class<?>... parameterTypes) {
        Method[] methods = Reflection.getMethods(clazz, Executor.class);
        List<Method> list = new ArrayList<Method>(methods.length);
        for (Method method : methods) {
            // Invariant: never null because "method" is only returned because it was annotated...
            Executor executor = method.getAnnotation(Executor.class);
            // Check if the executor is tagged with a context type matching the supplied context... 
            if (executor.value().isAssignableFrom(context)) {
                // Check formal parameter types of the method, ignoring primitives...
                if (Primitive.isEqualIgnorePrimitives(method.getParameterTypes(), parameterTypes)) {
                    list.add(method);
                }
            }
        }
        return list;
    }

    /**
      * Returns all declared constructors annotated with the {@link Executor} annotation 
      * in the class supplied as {@code clazz}, where the formal parameter types
      * match {@code parameterTypes} ignoring primitives, and where the 
      * {@link Executor#value() context} is assignable from {@code context}. 
      *  
      * @param  clazz          The class; cannot be null.
      * @param  context        The context; cannot be null.
      * @param  parameterTypes The formal parameter types for the methods; can be empty.
      * @return A list containing the proper annotated constructors in {@code clazz}, if
      *         any; never null, but may be empty.
      * @throws NullPointerException If {@code clazz} or {@code context} are null.         
      */
    public final static List<Constructor<?>> getConstructorExecutors(Class<?> clazz, Class<?> context, Class<?>... parameterTypes) {
        Constructor<?>[] constructors = Reflection.getConstructors(clazz, Executor.class);
        List<Constructor<?>> list = new ArrayList<Constructor<?>>(constructors.length);
        for (Constructor<?> constructor : constructors) {
            // Invariant: never null because "constructor" is only returned because it was annotated...
            Executor executor = constructor.getAnnotation(Executor.class);
            // Check if the executor is tagged with a context type matching the supplied context... 
            if (executor.value().isAssignableFrom(context)) {
                // Check formal parameter types of the method, ignoring primitives...
                if (Primitive.isEqualIgnorePrimitives(constructor.getParameterTypes(), parameterTypes)) {
                    list.add(constructor);
                }
            }
        }
        return list;
    }
        
}
