/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.reflect;

import java.lang.ref.WeakReference;

import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>weak reference entry</i> represents an object stored
  * as a weak reference, but where the {@link #hashCode()} and
  * {@link #equals(Object)} methods of the object are used over 
  * the identity comparison performed by {@link java.lang.ref.WeakReference
  * java.lang.ref.WeakReference}. <p>
  * 
  * Since stored as a weak reference, the wrapped object may be
  * garbage collected at any time. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * Delegation is favoured over inheritance of 
  * {@code java.lang.ref.WeakReference}. 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <T> The type of object stored as a weak reference by this entry.
  */
public class WeakReferenceEntry<T> {
    
    /**
      * The weak reference storing the object. <p>
      * 
      * Never null, but the actual object may be if it was 
      * garbage collected.
      */
    private final WeakReference<T> reference;
    /**
      * The hash code of the object stored as a weak reference from 
      * the time of creation of this entry. 
      */
    private final int hashCode;
   
     /**
       * Constructor.
       * 
       * @param  object The object to store as weak reference; cannot be null.
       * @throws NullPointerException If {@code object} is null. 
       */
    public WeakReferenceEntry(T object) {
        this.hashCode = object.hashCode();
        this.reference = new WeakReference<T>(object);
    }
   
    /**
      * Returns the object stored as a weak reference, or null if
      * garbage collected.
      *  
      * @return The object; or null if garbage collected.
      */
    public T get() {
        return this.reference.get();
    }
   
    /**
      * Returns the hash code of this entry. <p>
      * 
      * The hash code corresponds to the hash code of the object stored 
      * as a weak reference at the time of creation of this entry. 
      * 
      * @return The hash code.
      */
    public int hashCode() {
        return this.hashCode;
    }
   
    /**
      * Returns true if {@code object} represents the same object as this
      * entry, false if not. <p>
      * 
      * The test for object equality is performed using equals. <p>
      * 
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object object) {
        if (object == this) {
            return true;
        } else if (!(object instanceof WeakReferenceEntry)) {
            return false;
        }
        Object a = this.get();
        if (a == null) {
            return false;
        }
        return a.equals(((WeakReferenceEntry<?>)object).get());        
    }
    
    /**
      * Returns the string representation of this entry.
      *  
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.NAME_OPEN.toString(this);
        T object = this.get();
        if (object != null) {
            sb.append(object);
        }
        return sb.append(']').toString();
    }
   
}
