/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.util;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

/**
  * Utilities to <i>manipulate throwable types</i> and <i>objects</i>.
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class Throwables {

    /**
      * Private off-limit constructor.  
      */
    private Throwables() {
        throw new AssertionError(); // guard against reflection...
    }
    
    /**
      * Returns the first throwable assignable to {@code type} in the
      * cause chain acquired from {@code throwable}, if any. <p>
      * 
      * The returned throwable may be {@code throwable} it self.
      * 
      * @param  <T>       The type of throwable to find in the cause stack.
      * @param  throwable The throwable to supply the cause chain; cannot be null.
      * @param  type      The type as a class token; cannot be null.
      * @return The first throwable assignable to {@code type}, or null
      *         if no such throwable is found.
      * @throws NullPointerException If either argument is null.
      */
    public final static <T extends Throwable> T unwind(Throwable throwable, Class<T> type) {
        if (type.isAssignableFrom(throwable.getClass())) {
            return type.cast(throwable);
        }
        Throwable t = throwable;
        while ((t = t.getCause()) != null) {
            if (type.isAssignableFrom(t.getClass())) {
                return type.cast(t);
            }
        }
        return null;
    }

    /**
      * Returns a throwable of the type supplied as {@code type}
      * based on {@code throwable}. <p>
      * 
      * If {@code throwable} is already an instance of 
      * the desired type, it will simply be returned. Otherwise a new 
      * throwable of the desired type is created, {@code throwable} 
      * is set as the cause, and the created throwable is returned. <p>
      * 
      * This method can be used as a simple idiom for re-throwing the
      * correct type of exception if the exception cannot be handled, 
      * without having to use several different catch-blocks to avoid 
      * hiding the actual exception type. <p>
      * 
      * Consider:
      * <pre>
      * public foo() throws FooException {
      *   try {
      *     ..
      * 
      *   } catch (FooException fe) {
      *     throw fe;
      *   } catch (Exception e) {
      *     throw new FooException(e);
      *   }
      * }
      * </pre>
      * Using this method instead:
      * <pre>
      * public foo() throws FooException {
      *   try {
      *     ..
      * 
      *   } catch (Exception e) {
      *     throw toThrowableType(e, FooException.class);
      *   }
      * }
      * </pre>
      *   
      * @param  <T>       The type of throwable.
      * @param  throwable The throwable; cannot be null.
      * @param  type      The specific type as a class token; cannot be null.
      * @return A throwable having the desired type; never null.
      * @throws NullPointerException If either parameter is null.
      * @throws RuntimeException If the current context does not have
      *         access to create a throwable exception of the desired type; 
      *         or if the desired type does not have a no-arg constructor, 
      *         or the construction fails.
      * @see    #toThrowableType(Throwable, Class, boolean)        
      */
    public final static <T extends Throwable> T toThrowableType(Throwable throwable, Class<T> type) {
        return toThrowableType(throwable, type, false);
    }

    /**
      * Returns a throwable of the type supplied as {@code type}
      * based on {@code throwable}. <p>
      * 
      * If {@code throwable} is already an instance of the desired type, 
      * it will simply be returned. If not and if {@code unwind}
      * is true, a throwable of the desired type is tried unwound
      * from the cause chain acquired from {@code throwable} and if found
      * returned. Otherwise a new throwable of the desired type is created, 
      * {@code throwable} is set as the cause, and the created throwable 
      * is returned. <p>
      * 
      * This method can be used as a simple idiom for re-throwing the
      * correct type of exception if the exception cannot be handled, 
      * without having to use several different catch-blocks to avoid 
      * hiding the actual exception type. <p>
      * 
      * Consider:
      * <pre>
      * public foo() throws FooException {
      *   try {
      *     ..
      * 
      *   } catch (FooException fe) {
      *     throw fe;
      *   } catch (Exception e) {
      *     throw new FooException(e);
      *   }
      * }
      * </pre>
      * Using this method instead:
      * <pre>
      * public foo() throws FooException {
      *   try {
      *     ..
      * 
      *   } catch (Exception e) {
      *     throw toThrowableType(e, FooException.class, true); // true = try and unwind
      *   }
      * }
      * </pre>
      *   
      * @param  <T>       The type of throwable.
      * @param  throwable The throwable; cannot be null.
      * @param  type      The specific type as a class token; cannot be null.
      * @param  unwind    True to try and unwind a throwable of type {@code T}
      *                   from cause chain, false to create a new throwable
      *                   if needed (a new instance will also be created if
      *                   no such type can be found in the cause chain). 
      * @return A throwable having the desired type; never null.
      * @throws NullPointerException If {@code throwable} or {@code type} are null.
      * @throws RuntimeException If a throwable must be created and the current context 
      *         does not have access to create a throwable exception of the desired type; 
      *         or if the desired type does not have a no-arg constructor, 
      *         or the construction fails.
      * @see    #toThrowableType(Throwable, Class)        
      */
    public final static <T extends Throwable> T toThrowableType(Throwable throwable, Class<T> type, boolean unwind) {
        if (type.equals(throwable.getClass())) {
            return type.cast(throwable);
        }
        T exception = null;
        if (unwind) {
            exception = unwind(throwable, type);
        }
        if (exception == null) {
            try {
                exception = type.newInstance();
                exception.initCause(throwable);
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
        return exception;
    }
    
    /**
      * Returns a list containing the throwable cause chain acquired
      * from {@code throwable}. <p>
      * 
      * The first index contains {@code throwable}, the second index 
      * the cause A of {@code throwable}, the third the cause B 
      * of A, etc. 
      * 
      * @param  throwable The throwable to supply the cause chain; cannot be null.
      * @return The list containing the cause chain; never null, and never empty.
      * @throws NullPointerException If {@code throwable} is null.         
      */
    public final static List<Throwable> getCauses(Throwable throwable) {
        List<Throwable> causes = new ArrayList<Throwable>();
        Throwable t = throwable;
        causes.add(t);
        while ((t = t.getCause()) != null) {
            causes.add(t);
        }
        return causes;
    }
    
    /**
      * Returns a list containing the checked exceptions from {@code objects}. <p>
      * 
      * The returned list may contain duplicates.
      * 
      * @param  objects The objects; cannot be null, but may contain null entries.
      * @return A list containing the checked exceptions from
      *         {@code objects}; never null.
      * @throws NullPointerException If {@code objects} is null.          
      */
    public final static List<Exception> getCheckedExceptions(Object... objects) {
        return getCheckedExceptions(Arrays.asList(objects));
    }
    
    /**
      * Returns a list containing the checked exceptions from {@code objects}. <p>
      * 
      * The returned list may contain duplicates.
      * 
      * @param  objects The objects; cannot be null, but may contain null entries.
      * @return A list containing the checked exceptions from
      *         {@code objects}; never null.
      * @throws NullPointerException If {@code objects} is null.          
      */
    public final static List<Exception> getCheckedExceptions(Collection<?> objects) {
        List<Exception> checked = new ArrayList<Exception>(objects.size());
        for (Object object : objects) {
            if (object instanceof Exception) {
                if (!(object instanceof RuntimeException)) { 
                    checked.add(Exception.class.cast(object));
                }
            }
        }
        return checked;
    }

    /**
      * Returns a list containing the checked exception types from {@code types}. <p>
      * 
      * The returned list may contain duplicates.
      * 
      * @param  types The types; cannot be null, but may contain null entries.
      * @return A list containing the checked exception types from 
      *         {@code types}; never null. 
      * @throws NullPointerException If {@code types} is null.
      */
    public final static List<Class<? extends Exception>> getCheckedExceptionTypes(Class<?>... types) {
        return getCheckedExceptionTypes(Arrays.asList(types));
    }

    /**
      * Returns a list containing the checked exception types from {@code types}. <p>
      * 
      * The returned list may contain duplicates.
      * 
      * @param  types The types; cannot be null, but may contain null entries.
      * @return A list containing the checked exception types from 
      *         {@code types}; never null. 
      * @throws NullPointerException If {@code types} is null.
      */
    public final static List<Class<? extends Exception>> getCheckedExceptionTypes(Collection<Class<?>> types) {
        List<Class<? extends Exception>> checked = new ArrayList<Class<? extends Exception>>(types.size());
        for (Class<?> type : types) {
            if (Exception.class.isAssignableFrom(type)) {
                if (!(RuntimeException.class.isAssignableFrom(type))) { 
                    checked.add(Exception.class.cast(type).getClass());
                }
            }
        }
        return checked;
    }
    
    /**
      * Returns a list containing the unchecked throwables from {@code objects}. <p>
      * 
      * The returned list may contain duplicates.
      * 
      * @param  objects The objects; cannot be null, but may contain null entries.
      * @return A list containing the unchecked throwable types from
      *         {@code objects}; never null. 
      * @throws NullPointerException If {@code objects} is null.
      */
    public final static List<Throwable> getUncheckedThrowables(Object... objects) {
        return getUncheckedThrowables(Arrays.asList(objects));
    }

    /**
      * Returns a list containing the unchecked throwables from {@code objects}. <p>
      * 
      * The returned list may contain duplicates.
      * 
      * @param  objects The objects; cannot be null, but may contain null entries.
      * @return A list containing the unchecked throwable types from
      *         {@code objects}; never null. 
      * @throws NullPointerException If {@code objects} is null.
      */
    public final static List<Throwable> getUncheckedThrowables(Collection<?> objects) {
        List<Throwable> unchecked = new ArrayList<Throwable>(objects.size());
        for (Object object : objects) {
            if (object instanceof Throwable) {
                if ((object instanceof RuntimeException) || (!(object instanceof Exception))) {
                    unchecked.add(Throwable.class.cast(object));
                }
            }
        }
        return unchecked;
    }
    
    /**
      * Returns a list containing the unchecked throwable types from {@code types}. <p>
      * 
      * The returned list may contain duplicates.
      * 
      * @param  types The types; cannot be null, but may contain null entries.
      * @return A list containing the unchecked throwable types from
      *         {@code types}; never null. 
      * @throws NullPointerException If {@code types} is null.
      */
    public final static List<Class<? extends Throwable>> getUncheckedThrowableTypes(Class<?>... types) {
        return getUncheckedThrowableTypes(Arrays.asList(types));
    }
    
    /**
      * Returns a list containing the unchecked throwable types from {@code types}. <p>
      * 
      * The returned list may contain duplicates.
      * 
      * @param  types The types; cannot be null, but may contain null entries.
      * @return A list containing the unchecked throwable types from
      *         {@code types}; never null. 
      * @throws NullPointerException If {@code types} is null.
      */
    public final static List<Class<? extends Throwable>> getUncheckedThrowableTypes(Collection<Class<?>> types) {
        List<Class<? extends Throwable>> unchecked = new ArrayList<Class<? extends Throwable>>(types.size());
        for (Class<?> type : types) {
            if (Throwable.class.isAssignableFrom(type)) {
                if ((RuntimeException.class.isAssignableFrom(type)) || (!(Exception.class.isAssignableFrom(type)))) {
                    unchecked.add(Throwable.class.cast(type).getClass());
                }
            }
        }
        return unchecked;
    }
    
    /**
      * Returns a runtime exception based on {@code throwable}. <p>
      * 
      * If {@code throwable} is already a runtime exception, it will
      * simply be returned. Otherwise a new runtime exception is 
      * created, {@code throwable} is set as the cause, and the 
      * created runtime exception is returned.
      * 
      * @param  throwable The throwable; cannot be null.
      * @return A runtime exception; never null.
      * @throws NullPointerException If {@code throwable} is null.
      */
    public final static RuntimeException toRuntimeException(Throwable throwable) {
        return toRuntimeException(throwable, false);
    }

    /**
      * Returns a runtime exception based on {@code throwable}. <p>
      * 
      * If {@code throwable} is already a runtime exception, 
      * it will simply be returned. If not and if {@code unwind}
      * is true, a runtime exception is tried unwound from the cause chain 
      * acquired from {@code throwable} and if found
      * returned. Otherwise a new runtime exception is created, 
      * {@code throwable} is set as the cause, and the created runtime 
      * exception is returned. <p>
      * 
      * @param  throwable The throwable; cannot be null.
      * @param  unwind    True to try and unwind a runtime exception from cause 
      *                   chain, false to create a new runtime exception if needed 
      *                   (a new instance will also be created if no runtime 
      *                   exception can be found in the cause chain). 
      * @return A runtime exception; never null.
      * @throws NullPointerException If {@code throwable} is null.
      */
    public final static RuntimeException toRuntimeException(Throwable throwable, boolean unwind) {
        return toRuntimeException(throwable, RuntimeException.class, unwind);
    }

    /**
      * Returns a runtime exception of the type supplied as {@code type}
      * based on {@code throwable}. <p>
      * 
      * If {@code throwable} is already a runtime exception of 
      * the desired type, it will simply be returned. Otherwise a new 
      * runtime exception of the desired type is created, {@code throwable} 
      * is set as the cause, and the created runtime exception is returned.
      * 
      * @param  <T>       The type of throwable.
      * @param  throwable The throwable; cannot be null.
      * @param  type      The specific runtime exception type as a class token; 
      *                   cannot be null.
      * @return A runtime exception having the desired type; never null.
      * @throws NullPointerException If either parameter is null.
      * @throws RuntimeException If the current context to does have
      *         access to create a runtime exception of the desired type; 
      *         or if the desired type does not have a default constructor, 
      *         or the construction fails.
      */
    public final static <T extends RuntimeException> T toRuntimeException(Throwable throwable, Class<T> type)  {
        return toRuntimeException(throwable, type, false);
    }

    /**
      * Returns a runtime exception of the type supplied as {@code type}
      * based on {@code throwable}. <p>
      * 
      * If {@code throwable} is already a runtime exception of the desired type, 
      * it will simply be returned. If not and if {@code unwind} is true, a 
      * runtime exception of the desired type is tried unwound
      * from the cause chain acquired from {@code throwable} and if found
      * returned. Otherwise a new runtime exception of the desired type is created, 
      * {@code throwable} is set as the cause, and the created runtime 
      * exception is returned. <p>
      * 
      * @param  <T>       The type of throwable.
      * @param  throwable The throwable; cannot be null.
      * @param  type      The specific runtime exception type as a class token; 
      *                   cannot be null.
      * @param  unwind    True to try and unwind a runtime exception of the desired
      *                   type from cause chain, false to create such a new runtime exception 
      *                   if needed (a new instance will also be created if no such runtime 
      *                   exception can be found in the cause chain). 
      * @return A runtime exception having the desired type; never null.
      * @throws NullPointerException If either parameter is null.
      * @throws RuntimeException If the current context to does have
      *         access to create a runtime exception of the desired type; 
      *         or if the desired type does not have a default constructor, 
      *         or the construction fails.
      */
    public final static <T extends RuntimeException> T toRuntimeException(Throwable throwable, Class<T> type, boolean unwind)  {
        return toThrowableType(throwable, type, unwind);
    }

    /**
      * Returns true if {@code throwable} is a checked exception, false if not.
      *  
      * @param  throwable The throwable to test; can be null, in which case false is returned.
      * @return True if {@code throwable} is a checked exception, false if not.
      */
    public final static boolean isChecked(Throwable throwable) {
        return ((throwable instanceof Exception) && (!(throwable instanceof RuntimeException)));
    }
    
    /**
      * Returns a checked exception based on {@code throwable}. <p>
      * 
      * If {@code throwable} is already a checked exception, it will
      * simply be returned. Otherwise a new checked exception is 
      * created, {@code throwable} is set as the cause, and the 
      * created checked exception is returned.
      * 
      * @param  throwable The throwable; cannot be null.
      * @return A checked exception; never null.
      * @throws NullPointerException If {@code throwable} is null.
      */
    public final static Exception toCheckedException(Throwable throwable) {
        return toCheckedException(throwable, false);
    }

    /**
      * Returns a checked exception based on {@code throwable}. <p>
      * 
      * If {@code throwable} is already a checked exception, 
      * it will simply be returned. If not and if {@code unwind}
      * is true, a checked exception is tried unwound from the cause chain 
      * acquired from {@code throwable} and if found
      * returned. Otherwise a new checked exception is created, 
      * {@code throwable} is set as the cause, and the created checked
      * exception is returned. <p>
      * 
      * @param  throwable The throwable; cannot be null.
      * @param  unwind    True to try and unwind a checked exception from cause 
      *                   chain, false to create a new checked exception if needed 
      *                   (a new instance will also be created if no checked 
      *                   exception can be found in the cause chain). 
      * @return A checked exception; never null.
      * @throws NullPointerException If {@code throwable} is null.
      */
    public final static Exception toCheckedException(Throwable throwable, boolean unwind) {
        if (isChecked(throwable)) {
            return (Exception)throwable;
            
        } else if (unwind) {
            List<Exception> exceptions = getCheckedExceptions(getCauses(throwable));
            if (exceptions.size() > 0) {
                return exceptions.get(0);
            }
        }
        return new Exception(throwable);
    }

    /**
      * Returns a checked exception of the type supplied as {@code type}
      * based on {@code throwable}. <p>
      * 
      * If {@code throwable} is already a checked exception of 
      * the desired type, it will simply be returned. Otherwise a new 
      * checked exception of the desired type is created, {@code throwable} 
      * is set as the cause, and the created checked exception is returned.
      * 
      * @param  <T>       The type of throwable.
      * @param  throwable The throwable; cannot be null.
      * @param  type      The specific checked exception type as a class token; 
      *                   cannot be null.
      * @return A checked exception having the desired type; never null.
      * @throws NullPointerException If either parameter is null.
      * @throws RuntimeException If the current context to does have
      *         access to create a checked exception of the desired type; 
      *         or if the desired type does not have a default constructor, 
      *         or the construction fails.
      */
    public final static <T extends Exception> T toCheckedException(Throwable throwable, Class<T> type) {
        return toCheckedException(throwable, type, false);
    }
    
    /**
      * Returns a checked exception of the type supplied as {@code type}
      * based on {@code throwable}. <p>
      * 
      * If {@code throwable} is already a checked exception of the desired type, 
      * it will simply be returned. If not and if {@code unwind} is true, a 
      * checked exception of the desired type is tried unwound
      * from the cause chain acquired from {@code throwable} and if found
      * returned. Otherwise a new checked exception of the desired type is created, 
      * {@code throwable} is set as the cause, and the created checked 
      * exception is returned. <p>
      * 
      * @param  <T>       The type of throwable.
      * @param  throwable The throwable; cannot be null.
      * @param  type      The specific runtime exception type as a class token; 
      *                   cannot be null.
      * @param  unwind    True to try and unwind a checked exception of the desired
      *                   type from cause chain, false to create such a new checked exception 
      *                   if needed (a new instance will also be created if no such checked 
      *                   exception can be found in the cause chain). 
      * @return A checked exception having the desired type; never null.
      * @throws NullPointerException If either parameter is null.
      * @throws RuntimeException If the current context to does have
      *         access to create a checked exception of the desired type; 
      *         or if the desired type does not have a default constructor, 
      *         or the construction fails.
      */
    public final static <T extends Exception> T toCheckedException(Throwable throwable, Class<T> type, boolean unwind) {
        if (RuntimeException.class.isAssignableFrom(type)) {
            throw new ClassCastException("Type cannot be a runtime exception: " + type.getName());
        }
        return toThrowableType(throwable, type, unwind);
    }
    
    /**
      * Returns the last index + 1 in the supplied stack trace where one 
      * of the supplied tokens is found. <p>
      * 
      * The search for a token is done in the class name available via
      * a stack trace element using 
      * {@link Strings#indexOfIgnoreCase(CharSequence, CharSequence)}. 
      * Hence, the search is case-insensitive, and the search for tokens 
      * is done in the order the tokens are supplied.
      *
      * @param  trace  The stack trace; cannot be null.
      * @param  tokens The list of tokens to search for.
      * @return The index of the first token found, or {@code trace.length} 
      *         if no token could be found.
      * @throws NullPointerException If {@code trace} is null.
      */
    public final static int getIndex(StackTraceElement[] trace, CharSequence... tokens) {
        if (tokens.length == 0) {
            return trace.length; // all
        }
        for (int i = trace.length - 1; i >= 0; i--) {
            String name = trace[i].getClassName();
            for (CharSequence token : tokens) {
                if (Strings.indexOfIgnoreCase(name, token) != -1) {
                    return i + 1;
                }
            }
        }
        return trace.length; // all
    }
    
    /**
      * Serializes the stack trace from {@code throwable} into
      * the returned char sequence. <p>
      * 
      * Instead of writing to, for example, {@code System.err},
      * the same output is simply written to the returned char 
      * sequence, including newlines.
      *
      * @param  throwable The throwable; cannot be null.
      * @return A char sequence containing the serialized stack trace 
      *         from {@code throwable}; never null.
      * @throws NullPointerException If {@code throwable} is null.
      */
    public final static CharSequence toStackTrace(Throwable throwable) {
        StringWriter writer = new StringWriter();
        throwable.printStackTrace(new PrintWriter(writer));
        return writer.getBuffer();
    }
            
}
