/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.observer;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.model.Sequence.State;
import dk.rode.thesis.strategy.ObjectPolicy;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * A <i>correlated sequence observer</i> correlates the behaviour 
  * of an {@link ObservableSequence observable} sequence with another
  * (perhaps observable) {@link Sequence}. <p>
  * 
  * When {@link Sequence#next()} is invoked on the observed sequence, 
  * {@code next} is invoked on the correlated sequence. The same is 
  * true for {@link Sequence#reset()}, either if implicit in form of 
  * a bounded sequence that wraps, or because the {@code reset()}
  * method was explicitly invoked. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("ConcreteObserver")
public class CorrelatedSequenceObserver implements SequenceObserver<State> {

    /**
      * The correlated sequence. <p>
      * 
      * Never null.
      */
    private final Sequence<?> sequence;
    
    /**
      * Constructor.
      * 
      * @param  sequence The sequence to correlate; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public CorrelatedSequenceObserver(Sequence<?> sequence) {
        if ((this.sequence = sequence) == null) {
            throw new NullPointerException("Correlated sequence cannot be null");
        }
    }

    /**
      * 
      */
    public void sequenceEvent(Sequence<?> sequence, State state) {
        switch (state) {
            case NORMAL: {
                this.sequence.next();
                break;
            }
            case RESET:   
            case RESTART: {
                this.sequence.reset();
                break;
            }
            default:
                return;
        }
        LogFactory.getLog(this).println("-> " + this.sequence);
    }

    /**
      * Returns the string representation of this observer. 
      *
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.NAME_OPEN.toString(this);
        return sb.append(this.sequence.toString(SequencePolicy.IDENTITY)).append(']').toString();
    }
    
}
