/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.observer;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.model.Sequence.State;
import dk.rode.thesis.meta.reflect.Executor;
import dk.rode.thesis.strategy.ObjectPolicy;
import dk.rode.thesis.strategy.StringablePolicy;

/**
  * A <i>print sequence observer</i> prints the state of any 
  * {@link ObservableSequence observable} sequence it is
  * observing. <p>   
  *
  * <b>Implementation notes</b>:<br>
  * The {@link #sequenceEvent(Sequence, Sequence.State)} method
  * is also annotated with the {@link Executor} annotation and
  * instances of this class can thus also be observers for 
  * {@link AnnotatedObserversSequence} sub-classes or used
  * directly by the {@link ObserverManager} class. 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("ConcreteObserver")
public class PrintSequenceObserver implements SequenceObserver<Sequence.State> {

    /**
      * The policy to use to print the observable sequence. <p>
      * 
      * Can be null, i.e. default policy.
      */
    private final StringablePolicy<Sequence<?>> policy;
    
    /**
      * Constructor. <p>
      * 
      * The default policy for the observable sequence is used.
      */
    public PrintSequenceObserver() {
        this(null);
    }
    
    /**
      * Constructor.
      * 
      * @param  policy The policy to use; can be null, i.e. default policy.
      * @throws NullPointerException If {@code out} is null.
      */
    public PrintSequenceObserver(StringablePolicy<Sequence<?>> policy) {
        this.policy = policy;
    }

    /**
      * {@inheritDoc}
      * 
      * This method is annotated with the {@link Executor} annotation
      * and can thus also be used to notify instances of this class using 
      * any implementation of {@link AnnotatedObserversSequence} as well.
      */
    @Executor
    public void sequenceEvent(Sequence<?> sequence, State state) {
        Log out = LogFactory.getLog(this);
        out.println("Sequence has changed state: " + sequence.toString(this.policy) + ": " + state);
        // Specific action based on supplied state...
        switch (state) {
            case RESTART: {
                out.println("Sequence has been restarted!");
                break;
            }
            case RESET: {
                out.println("Sequence has been reset!");
                break;
            }
        }
    }

    /**
      * Returns the string representation of this observer. 
      *
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.NAME_OPEN.toString(this);
        return sb.append(this.policy).append(']').toString();
    }
    
}
