/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */

/**
  * Implementations and examples of the <b>Observer</b> 
  * design pattern [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.293]. <p>
  *  
  * <b>Intent</b>:<ul><p><i>
  * 
  *    Define a onetomany dependency between objects so that 
  *    when one object changes state, all dependants are notified 
  *    and updated automatically.
  *    
  * </i></ul>
  * 
  * Here, the <b>Subject</b> participant is represented by the
  * {@link dk.rode.thesis.observer.Observable} interface. The 
  * {@code Observable} interface only defines methods for 
  * <i>adding</i>, <i>removing</i>, and <i>fetching</i> observers, 
  * but no a method to <i>notify</i> the observers. The
  * type of observers must also be specified by a type parameter. 
  * All this is left up to the actual implementations, because 
  * different implementations will have different needs to the type of 
  * notification performed, for example in the type of arguments 
  * used, the notification order, asynchronous notification, etc. 
  * The {@link dk.rode.thesis.observer.AspectObservable} type extends
  * {@code Observable} to allow observers to subscribe to specific
  * <i>aspects</i> only. <p> 
  * 
  * A refinement of the {@code Observable} interface is the 
  * {@link dk.rode.thesis.observer.ObservableSequence}
  * type, which represents a {@link dk.rode.thesis.meta.model.Sequence}
  * that will notify observers when its internal 
  * aspect changes, typically when its internal 
  * {@link dk.rode.thesis.meta.model.Sequence.State state} has changed. 
  * The {@link dk.rode.thesis.observer.AspectObservableSequence} sub-type
  * allows observers to sub-scribe to specific aspects only, i.e. to 
  * specific internal state values only. <p>
  *   
  * For observable sequences, the <b>Observer</b> participant is 
  * partly represented by the {@link dk.rode.thesis.observer.SequenceObserver} 
  * interface, which defines the notification method to use.
  * As a feature, {@code ObserverableSequence} 
  * extends {@code SequenceObserver}, and thus uses the inherited 
  * notification method to notify its observers. If the actual
  * {@code ObservableSequence} at the same time chooses to
  * use {@code SequenceObserver} as the observer type, the
  * notification method is used to both <i>issue</i> and 
  * <i>receive</i> notifications. This makes the observable
  * {@linkplain dk.rode.thesis.composite composite}. <p>  
  * 
  * However, observers need not be related to sequences, nor be required 
  * to use the {@code SequenceObserver} interface. As an alternative, 
  * any object annotated with the {@link dk.rode.thesis.meta.reflect.Executor} 
  * annotation may be considered a representative of the <b>Observer</b> 
  * participant under certain conditions (see implementation notes below). <p>  
  * 
  * Two abstract sub-classes of the {@code ObservableSequence} interface 
  * have been defined as well, also representing the <b>Subject</b> 
  * participant, namely the {@link dk.rode.thesis.observer.SequenceObserversSequence}
  * and {@link dk.rode.thesis.observer.AnnotatedObserversSequence} classes. 
  * The first represents an observable sequence that uses 
  * {@code SequenceObserver} types as observers, while the latter represents
  * an observable sequence that uses <i>annotated objects</i> as observers. <p>
  * 
  * The <b>ConcreteSubject</b> participant is represented by any 
  * non-abstract implementation of the {@code Observable} interface. The
  * {@link dk.rode.thesis.observer.DateSequence} class is an observable
  * sequence inheriting {@code SequenceObserversSequence}, while the
  * {@link dk.rode.thesis.observer.ObserverManager} class is a stand-alone
  * observer manager storing annotated objects as observers; it is not
  * specifically related to sequences and the changed subjects are supplied
  * as arguments to the notification method (<i>Change Manager</i>). A 
  * specific implementation of the {@code AnnotatedObserversSequence} 
  * class have been made in form of the  
  * {@link dk.rode.thesis.observer.AnnotatedObserversSequenceDecorator}
  * class that can be used to make any sequence observable using annotated
  * objects as observers. <p> 
  * 
  * The {@link dk.rode.thesis.observer.CorrelatedSequenceObserver},  
  * {@link dk.rode.thesis.observer.PrintSequenceObserver}, 
  * {@link dk.rode.thesis.observer.ProbeSequenceObserver}, 
  * and {@link dk.rode.thesis.observer.BirthdayRegistry}
  * classes all represent the <b>ConcreteObserver</b> participant. 
  * All except {@code ProbeSequenceObserver} implements the 
  * {@code SequenceObserver} interface. The latter uses annotations and
  * so does the {@code PrintSequenceObserver} class. The 
  * {@code PrintSequenceObserver} can thus be used as an observer for
  * both abstract sub-classes of {@code ObserverableSequence}. <p>
  * 
  * <a name="uml"></a><b>UML Class Diagram</b>:<br>
  * <img src="doc-files/Observer.png" border="0"><p>
  * 
  * <b>Implementation notes</b>:<br/>
  * The {@link dk.rode.thesis.observer.AnnotatedObserversSequence} class 
  * internally uses a {@link dk.rode.thesis.observer.ObserverManager} 
  * via object composition. Annotated observers works by annotating one or
  * more methods in a given observer class - regardless of visibility, name, 
  * and if static - with the {@link dk.rode.thesis.meta.reflect.Executor} 
  * annotation. However, the formal parameter types of the methods must 
  * match those supplied at construction time to the 
  * {@code AnnotatedObserversSequence} instance used (and thus to the 
  * {@code ObserverManager} used). At notification time, corresponding
  * arguments must then supplied to the notification method that reflectively
  * will invoke each annotated method. <p>
  * 
  * The {@code BirthdayRegistry} class uses an inner class to implement 
  * the observer functionality, which ensures that we do not expose the 
  * {@link dk.rode.thesis.observer.SequenceObserver} interface for misuse. 
  * This is common practice in Java. However, since the inner class is in 
  * effect an {@linkplain dk.rode.thesis.adapter adapter}, the exact type 
  * of notification mechanism used by the observable sequence must 
  * be known. For the birthday registry, the {@code SequenceObserver} 
  * interface is used, and this limits the type of applicable observable 
  * sequences to {@code SequenceObserver<SequenceObserver,Date>}. <p>
  *    
  * As a curiosity, Java provides the {@link java.util.Observable java.util.Observable} 
  * and {@link java.util.Observer java.util.Observer} classes to support the 
  * Observer pattern, and has done so since version 1.0 or so. To our knowledge, 
  * these classes are rarely used. The interface they provide seems far from
  * adequate, and most methods are synchronised. Internally a 
  * {@link java.util.Vector java.util.Vector} is even used, practically 
  * deprecated since version 1.2!
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a> 
  */
@Pattern(name         = "Observer", 
         scope        = Scope.Object,
         purpose      = Purpose.Behavioural,
         participants = {"Subject", "ConcreteSubject", "Observer", "ConcreteObserver"})
package dk.rode.thesis.observer;

import dk.rode.thesis.meta.Pattern;
import dk.rode.thesis.meta.Purpose;
import dk.rode.thesis.meta.Scope;

