/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.singleton;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.ArraySequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.StringablePolicy;

/**
  * A <i>singleton alphabet sequence</i> for letters in the Danish alphabet: 
  * {@code a} though {@code z} as well as {@code }, {@code }, and
  * {@code }, all lower-cased. <p>
  * 
  * This class cannot be sub-classed, but <b>is</b> serializable. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * By using an enumeration with a single {@link #Instance constant} to represent 
  * the singleton instance of this sequence, many singleton traits are enforced by
  * Java: the constructor has to be private and cannot be instantiated from elsewhere,
  * not even via reflection. The sequence type is guaranteed not cloneable, but 
  * is serializable! Lazy loading is also possible. There is no need to implement 
  * {@code hashCode()} and {@code equals(Object)}, as no other sequence will be 
  * identical. Identity (==) comparison can be used. The singleton instance can 
  * still be referenced through a static singleton method, but also through the 
  * constant. We consider a Singleton implementation using enumerations in this manner 
  * for an application of the 
  * <a href="http://www.rode.dk/thesis/ComparativeEvaluation.htm#section-7.1.1.4" target="thesis"><i>Singleton-as-Single-Constant</i></a> 
  * idiom. <p>
  * 
  * Passing arguments to a singleton type is difficult, though. The type cannot
  * inherit functionality as the {@link java.lang.Enum java.lang.Enum} class
  * is already inherited. Hence, composition has to be used in combination with
  * interface implementation. <p>
  * 
  * The inherited static method {@link #valueOf(String)} from {@code java.lang.Enum}
  * can also be used to initialise and/or fetch the singleton instance. This is 
  * utilised here to be able to fetch this sequence from a 
  * {@link StatelessSingletonRegistry} instance as it is annotated 
  * with the {@link Singleton} enumeration.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    NorwegianAlphabetSequence
  * @see    SimpsonsFamilySequence
  * @see    SimpsonsAndBouvierFamilySequence
  * @see    SmileySequence
  */
@Participant("Singleton")
@Singleton(value = "valueOf", arguments = "Instance")
public enum DanishAlphabetSequence implements Sequence<String> {

    /**
      * The {@link DanishAlphabetSequence} singleton instance
      * to use. <p>
      * 
      * Is not initialised until first time accessed. 
      */
    Instance;
    
    /**
      * Delegate sequence storing the actual letters in the
      * Danish alphabet. <p>
      * 
      * <b>Implementation notes</b>:<br>
      * Delegation has to be used as enumerations cannot extend other
      * classes as they already extend the {@link java.lang.Enum java.lang.Enum}
      * class. <p>
      * 
      * This cannot be declared static because then it might be loaded <b>after</b>
      * {@link #Instance}, as {@code Instance} is in fact a separate class. As
      * only a single constant is ever created, this works just fine.
      */
    private final Sequence<String> sequence = new ArraySequence<String>(
        "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", 
        "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z",
        "", "", "") {
        
        // Could ignore, as this method will never be invoked...
        public Sequence<String> copy() {
            return Instance;
        }
    };

    /**
      * The size of the Danish alphabet. <p>
      *  
      * <b>Implementation notes</b>:<br>
      * Accessing this static field will <b>before</b> {@link #Instance}
      * will <b>not</b> cause the singleton instance to be created!
      * 
      * @see NorwegianAlphabetSequence#SIZE
      * @see SimpsonsFamilySequence#SIZE
      * @see SimpsonsAndBouvierFamilySequence#SIZE
      * @see SmileySequence#SIZE
      */
    public final static int SIZE = 29; // has to be constant as not to force loading of Instance...
    
    /**
      * Private no-arg singleton constructor. <p>
      * 
      * <b>Implementation notes</b>:<br>
      * Java requires enumeration constructors to be private.
      */
    private DanishAlphabetSequence() {
        LogFactory.getLog(this).println("Created singleton: ", this);
    }
    
    /**
      * 
      */
    public boolean bounded() {
        return true;
    }

    /**
      * 
      */
    public boolean consistent() {
        return true;
    }

    /**
      * 
      */
    public boolean unique() {
        return true;
    }
    
    /**
      * Returns this instance.
      * 
      * @return This instance; never null.
      */
    public DanishAlphabetSequence copy() {
        return this;
    }

    /**
      * 
      */
    public String current() {
        return this.sequence.current();
    }

    /**
      * 
      */
    public String next() {
        return this.sequence.next();
    }

    /**
      * 
      */
    public void reset() {
        this.sequence.reset();
    }

    /**
      * 
      */
    public Sequence.State state() {
        return this.sequence.state();
    }

    /**
      * 
      */
    public String toString() {
        return this.toString(null).toString();
    }
    
    /**
      * 
      */
    public StringablePolicy<? super Sequence<String>> getStringablePolicy(StringablePolicy<? super Sequence<String>> policy) {
        return this.sequence.getStringablePolicy(policy);
    }

    /**
      * 
      */
    public CharSequence toString(StringablePolicy<? super Sequence<String>> policy) {
        return new StringBuilder().append("DanishAlphabetSequence").append(this.sequence.toString(policy)).toString();
    }
        
}
