/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.singleton;

/**
  * A <i>singleton registry</i> is a container for singleton 
  * instances of any type, {@code S}, inheriting the common
  * super type, {@code T}. <p>
  * 
  * A singleton type is identified by a <i>class token</i> supplied
  * to the {@link #getInstance(Class)} method, but how the actual
  * singleton creation and/or delivery is made is decided by the
  * actual registry implementation. <p>
  * 
  * The registry also decides the specific singleton types 
  * {@link #isSingleton(Class) supported}. <p>
  * 
  * Registries do <b>not</b> override {@code hashCode} or {@code equals(Object)}
  * and are therefore only considered equal based on identity. <p>
  *   
  * <b>Implementation notes</b>:<br>
  * Registry functionality is completely separated from singleton types/instances. <p>
  * 
  * Unlike the registry functionality described by Gamma et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.131],
  * a {@code SingletonRegistry} instance will never return a null value
  * in case a singleton type cannot be found. The calling context should 
  * never have to test for a null return value, because either the singleton 
  * instance exist (or be created if need be) or it does not; a null value 
  * does not make any sense, and would in our view constitute a poor design
  * decision considering the intrinsic conceptual behaviour in the Singleton 
  * Pattern. Hence, in the general case, the context should not have to worry 
  * about checked exceptions as they would never be thrown. For convenience, 
  * the {@link SingletonRegistry#isSingleton(Class)} method can be used to
  * test if a given type is a singleton supported by the registry. <p> 
  * 
  * By declaring an interface to implement for required registry functionality,
  * as opposed to using static methods directly, we get all the standard
  * benefits of object behaviour and polymorphism at the cost of having 
  * to create (possibly singleton) registries as well. For example,
  * a registry implementation may be synchronised or otherwise proxied
  * or decorated, or used as a private member variable in a factory. <p> 
  * 
  * The upper bound {@code T} on type parameters for implementation 
  * classes <i>can</i> be used to ensure that all singleton types for
  * a given registry must implement some sort of {@code Singleton}
  * interface, if so desired.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <T> The super type of all singleton types handled by
  *             this registry. Use {@code Object} if any type
  *             of singleton type can be handled.
  */
public interface SingletonRegistry<T> {
    
    /**
      * Returns the singleton instance of the type supplied as 
      * {@code type}. <p>
      * 
      * Whether or not {@code type} is a singleton type handled
      * by this registry can be determined by the {@link #isSingleton(Class)}
      * method.
      *  
      * @param  <S>  The actual singleton type.
      * @param  type The class token identifying the singleton type; cannot be null.
      * @return The singleton instance of the type supplied as {@code type};
      *         never null. 
      * @throws NullPointerException If {@code type} is null.
      * @throws SingletonException If {@code type} is not a singleton type
      *         handled by this registry, or if the type cannot be created when
      *         first requested.
      */
    public <S extends T> S getInstance(Class<S> type);
    
    /**
      * Returns true of the type supplied as {@code type} is a singleton 
      * type handled by this registry, false if not. <p>
      * 
      * Even if this method returns true, the actual singleton instance
      * may not be created before it is {@link #getInstance(Class) requested}.
      * The creation of the singleton instance can therefore still
      * fail!
      * 
      * @param  type The class token identifying the type; cannot be null.
      * @return True if {@code type} represents a singleton type handled by
      *         this registry. 
      * @throws NullPointerException If {@code type} is null.
      */
    public boolean isSingleton(Class<?> type);   
 
}
