/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.singleton;

import java.util.Arrays;
import java.util.Collections;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.ArraySequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.reflect.CallerClass;

/**
  * A <i>smiley sequence</i> is a sequence delivering different 
  * smileys, e.g. :), *<;), :(, etc. <p>
  * 
  * A smiley sequence is {@link Sequence#bounded() bounded}, 
  * {@link Sequence#consistent() consistent}, and
  * {@link Sequence#unique() unique} with respect to its
  * lifetime; the sequence of the generated smileys is determined
  * randomly at construction time. <p>
  *  
  * <b>Implementation notes</b>:<br>
  * The {@code SmileySequence} uses double-checked locking to perform
  * lazy initialisation of the singleton instance when first requested. <p> 
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    DanishAlphabetSequence
  * @see    NorwegianAlphabetSequence
  * @see    SimpsonsFamilySequence
  * @see    SimpsonsAndBouvierFamilySequence
  */
@Participant("Singleton")
@Singleton("getSequence")
public final class SmileySequence extends ArraySequence<String> {

    /**
      * Single instance of the sequence, lazily initialised using double-checked
      * locking when first requested. <p>
      * 
      * Will be null until initialised.
      */
    private static volatile SmileySequence instance; // must be volatile for double-checked locking to work!

    /**
      * All generated smileys. <p>
      */
    private final static String[] smileys = {
        ":)", ";)", ":|", ">:)", ">;)", ">:|", ">;|", ">:P", ">;p",  "*<:)", "*<;)", ">8(",
        ":(", ";(", ":P", ";P",  ">:(", ">;(", ">:P", ">;P", "*<:(", "*<;(", ">8)",  ">8|" 
    };
    
    /**
      * The number of different smileys returned by the sequence. <p>
      *  
      * <b>Implementation notes</b>:<br>
      * Accessing this static field <b>before</b> {@link #getSequence()} 
      * will <b>not</b> cause the singleton instance to be created!
      * 
      * @see DanishAlphabetSequence#SIZE
      * @see NorwegianAlphabetSequence#SIZE
      * @see SimpsonsFamilySequence#SIZE
      * @see SimpsonsAndBouvierFamilySequence#SIZE
      */
    public final static int SIZE = smileys.length;
    
    /**
      * Returns the {@link SmileySequence} singleton
      * instance to use. <p>
      * 
      * If this is the first time the sequence is requested, 
      * it will be lazily initialised.
      *
      * @return The sequence; never null.
      * @see    Singleton
      */
    public final static SmileySequence getSequence() {
        // Invariant: "instance" must be declared volatile for this to work, 
        // and only in Java 5 or higher!
        SmileySequence sequence = instance;
        if (sequence == null) { // 1st check: no locking
            synchronized (SmileySequence.class) {
                if ((sequence = instance) == null) { // 2nd check: with locking
                    instance = sequence = new SmileySequence();
                }
            }
        }
        return sequence;
    }
    
    /**
      * Private off-limit constructor.
      */
    private SmileySequence() {
        super(shuffleSmileys());
        Log out = LogFactory.getLog(this);
        // Guard against reflection...
        CallerClass caller = new CallerClass();
        // Sufficient to test on the caller class...
        if (!caller.isThisClass()) {
            out.error("Illegal access to singleton constructor: ", caller);
            throw new SingletonError(caller);
        }
        out.println("Created singleton: ", this);
    }
    
    /**
      * Generates a random sequence of smileys to return by the smiley sequence. <p>
      *  
      * @return A sequence of smileys in form of a string array; never null.
      */
    private final static String[] shuffleSmileys() {
        // Also shuffles "smileys" array...
        Collections.shuffle(Arrays.asList(smileys));
        return smileys;        
    }
    
    /**
      * Returns this instance.
      * 
      * @return This instance; never null.
      */
    public SmileySequence copy() {
        return this;
    }
    
}
