/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.state;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.AbstractSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.reflect.Reflection;
import dk.rode.thesis.meta.reflect.proxy.ProxyFactory;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>step sequence</i> represents an unbounded integer value that 
  * is increased/decreased with a given <i>step</i> with each call to 
  * {@link #next()}. The step may be negative, but not zero. <p>
  * 
  * A step sequence is {@link Sequence#bounded() unbounded}, but
  * {@link #unique() unique} and {@link Sequence#consistent() consistent}. <p>
  * 
  * Step sequences are created by the {@link StepSequence.Factory#create(int, int)} 
  * factory method. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * A step sequence is implemented using dynamic proxies. It will appear to change 
  * its class (state) when the sequence value changes from <i>even</i> to <i>odd</i> 
  * or vice versa. This is done by altering the <i>target object</i> for the sequence 
  * methods in a dynamic proxy when the value "flips". Hence, the implementation of
  * the sequence methods in effect changes. This imitates dynamic inheritance as 
  * described by Gamma et al. as the means to change the internal state [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.309].
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("Context")
public interface StepSequence extends Sequence<Integer> {

    /**
      * A factory to create {@link StepSequence} instances.  
      *
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      */
    public final static class Factory {
        /**
          * The proxy factory used to create the dynamic proxies. <p>
          * 
          * Never null.
          */
        private final static ProxyFactory factory = new ProxyFactory();
        
        /**
          * Private off-limit constructor. 
          */
        private Factory() {
            throw new AssertionError(); // guard against reflection...
        }
        
        /**
          * Creates a new step sequence using a step of {@code step} and 
          * starting with the initial value supplied as {@code value}.
          *  
          * @param  step  The step; cannot be zero.
          * @param  value The initial sequence value.
          * @return A new {@link StepSequence} instance; never null.
          * @throws IllegalArgumentException If {@code step} is zero.
          */
        public final static StepSequence create(int step, int value) {
            return factory.getProxy((StepSequence)null, // NetBeans complains otherwise... 
                                    new Handler(step, value), 
                                    StepSequence.class);
        }
    }
    
}

/**
  * An abstract class representing the basic traits of the 
  * {@link StepSequence} interface. <p>
  * 
  * This class has two sub-classes: one that handles even sequence
  * values, and one that handles odd values. It is possible to "flip"
  * between the two types using the {@link #flip()} method. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * This class could implement almost all sequence functionality, but
  * that is left to the sub-classes to better illustrate that an
  * different target is used once flipped.  
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("State")
abstract class StepSequenceImpl extends AbstractSequence<Integer> implements StepSequence {

    /**
      * The step value to increase the sequence value with each 
      * call to {@code next()}. <p>
      * 
      * Never zero, but can be negative.
      */
    final int step;
    /**
      * The initial value of this sequence. 
      */
    private final int start;
    /**
      * The current value of this sequence.
      */
    int current;
    
    /**
      * Constructor.
      * 
      * @param  step  The step; cannot be zero.
      * @param  value The initial sequence value.
      * @throws IllegalArgumentException If {@code step} is zero.
      * @see    StepSequence.Factory
      */
    StepSequenceImpl(int step, int value) {
        if ((this.step = step) == 0) {
            throw new NullPointerException("Step cannot be zero");
        }
        this.start = this.current = value;
    }

    /**
      * Copy constructor.
      * 
      * @param  sequence The sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is zero.
      */
    StepSequenceImpl(StepSequenceImpl sequence) {
        super(sequence);
        this.start = sequence.start;
        this.step = sequence.step;
        this.current = sequence.current;
    }
    
    /**
      * 
      */
    public boolean bounded() {
        return false;
    }

    /**
      * 
      */
    public boolean consistent() {
        return true;
    }

    /**
      * 
      */
    public boolean unique() {
        return true;
    }

    /**
      * 
      */
    public void reset() {
        this.current = this.start;
        super.reset();
    }
    
    /**
      * Flips this sequence from even to odd or vice versa.
      * 
      * @return A flipped instance; never null.
      */
    abstract StepSequenceImpl flip();

}

/**
  * The <i>even sequence</i> represents the state generating 
  * even {@link StepSequence step} sequence values. 
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    OddSequence
  */
@Participant("ConcreteState")
class EvenSequence extends StepSequenceImpl {

    /**
      * Constructor.
      * 
      * @param  step  The step; cannot be zero.
      * @param  value The initial sequence value.
      * @throws IllegalArgumentException If {@code step} is zero.
      */
    EvenSequence(int step, int value) {
        super(step, value);
    }

    /**
      * Copy constructor.
      * 
      * @param sequence The sequence to copy; never null.
      */
    EvenSequence(StepSequenceImpl sequence) {
        super(sequence);
    }
    
    /**
      * 
      */
    public EvenSequence copy() {
        return new EvenSequence(this);
    }

    /**
      * 
      */
    public Integer current() {
        return this.current;
    }

    /**
      * 
      */
    public Integer next() {
        this.state = State.NORMAL;
        return this.current += this.step;
    }
 
    /**
      * 
      */
    StepSequenceImpl flip() {
        return new OddSequence(this);
    }
}

/**
  * The <i>odd sequence</i> represents the state generating 
  * odd {@link StepSequence step} sequence values. 
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    EvenSequence
  */
@Participant("ConcreteState")
class OddSequence extends StepSequenceImpl {

    /**
      * Constructor.
      * 
      * @param  step  The step; cannot be zero.
      * @param  value The initial sequence value.
      * @throws IllegalArgumentException If {@code step} is zero.
      */
    OddSequence(int step, int value) {
        super(step, value);
    }

    /**
      * Copy constructor.
      * 
      * @param sequence The sequence to copy; never null.
      */
    OddSequence(StepSequenceImpl sequence) {
        super(sequence);
    }

    /**
      * 
      */
    public OddSequence copy() {
        return new OddSequence(this);
    }

    /**
      * 
      */
    public Integer current() {
        return this.current;
    }

    /**
      * 
      */
    public Integer next() {
        this.state = State.NORMAL;
        return this.current += this.step;
    }

    /**
      * 
      */
    StepSequenceImpl flip() {
        return new EvenSequence(this);
    }
    
}

/**
  * The <i>handler</i> class is used to perform the actual state change
  * when the {@link StepSequence step} sequence value goes from even to 
  * odd or vice versa. 
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    StepSequence.Factory#create(int, int)
  */
class Handler implements InvocationHandler {
    
    /**
      * The current sequence target. <p>
      * 
      * This represents the current state. <p>
      * 
      * Never null.
      */
    private StepSequenceImpl sequence;
    
    /**
      * Constructor.
      * 
      * @param  step  The step; cannot be zero.
      * @param  value The initial sequence value.
      * @throws IllegalArgumentException If {@code step} is zero.
      */
    Handler(int step, int value) {
        this.sequence = (value % 2 == 0) ? new EvenSequence(step, value) : new OddSequence(step, value);
    }
 
    /**
      * 
      */
    public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
        boolean evenBefore = this.sequence.current() % 2 == 0;
        Object result = Reflection.invoke(this.sequence, method, args);
        boolean evenAfter = this.sequence.current() % 2 == 0;
        if (evenBefore != evenAfter) {
            this.sequence = this.sequence.flip(); // change implementation
            LogFactory.getLog(StepSequence.class).println("Changed implementation to: ", this.sequence);
        }
        return result;
    }
    
    /**
      * Returns the string representation of this handler.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.NAME_OPEN.toString(this);
        return sb.append("; state: ").append(this.sequence).append(']').toString();
    }
    
}        

