/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.templatemethod;

import java.io.File;
import java.io.FileFilter;
import java.io.InputStream;

import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.meta.util.Primitive;
import dk.rode.thesis.meta.util.Util;

/**
  * <b>Template Method</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class Main implements Testable {

    /**
      * Executes the <b>Template Method</b> tests.
      * 
      * @param  args The arguments to supply to {@link #test(Log, Arguments)},
      *              if any; can be null.
      * @throws Exception If the test fails unexpectedly.             
      */
    public static void main(String[] args) throws Exception {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }

    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * 
      */
    public boolean test(Log out, Arguments arguments) throws Exception {
        out.heading("Template Method tests...");
        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }

        boolean log = arguments.getArgumentValue("log", true, Boolean.class);

        LogFactory.getLog(ZipSequence.class).setActive(log);
        LogFactory.getLog(FileSequence.class).setActive(log);
        LogFactory.getLog(NegativeSequence.class).setActive(true);

        //
        // Created sequences...
        //
        
        // Simplest sequence template implementation...
        NegativeSequence ns = new NegativeSequence(-17);

        // Create a zip sequence fetching zip entries form zip file
        // in data directory...
        ZipSequence zs = new ZipSequence(Util.getDataDirectory(), "foobar.zip");
        
        // Create a file sequence fetching .txt files from the data directory...
        FileSequence fs = 
            new FileSequence(Util.getDataDirectory(), new FileFilter() {
                public boolean accept(File file) {
                    if (file.isDirectory()) {
                        return false;
                    }
                    return file.getName().endsWith(".txt");
                }
            });
        
        out.heading("Created sequences");
        out.println("Sequence 0: ", ns)
           .println("Sequence 1: ", zs)
           .println("Sequence 2: ", fs)
           .println();
        
        //
        // Sequence 0
        //
        
        out.heading("Sequence 0: ", ns);
        try {
            for (int i = 0; i < 10; i++) {
                Integer value = ns.current();
                out.println(i, ": value = ", value);
                ns.next();
            }
        } finally {
            ns.close();
        }

        //
        // Sequence 1
        //
        
        out.println().heading("Sequence 1: ", zs);
        try {
            for (int i = 0; i < 10; i++) {
                InputStream value = zs.current();
                byte[] bytes = Util.readBytes(value);
                out.println(i, ": value = ", value)
                   .println(i, ": as byte[]: ", Primitive.toString(bytes))
                   .println(i, ": as String:\n", new String(bytes, "UTF-8"));
                zs.next();
                
                if (i % 3 == 0) {
                    value.close(); // is not allowed!
                } else if (i % 8 == 0) {
                    value.reset(); // is not allowed!
                }
            }
        } finally {
            zs.close();
        }

        //
        // Sequence 2
        //
        
        out.println().heading("Sequence 2: ", fs);
        try {
            for (int i = 0; i < 10; i++) {
                byte[] value = fs.current();
                out.println(i, ": value = ", Primitive.toString(value))
                   .println(i, ": as String:\n", new String(value, "UTF-8"));
                fs.next();
            }
        } finally {
            fs.close();
        }
        
        out.heading("Template Method succeeded!");
        
        return true;
    }
       
}
