/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.visitor;

import java.util.Date;

import dk.rode.thesis.adapter.AdapterStrategy;
import dk.rode.thesis.adapter.SequenceAdapter;
import dk.rode.thesis.composite.CharSequenceCompositeSequence;
import dk.rode.thesis.composite.CompositePolicy;
import dk.rode.thesis.composite.CompositeSequence;
import dk.rode.thesis.composite.CompositeStrategy;
import dk.rode.thesis.facade.RandomSequence;
import dk.rode.thesis.flyweight.CharacterFactory;
import dk.rode.thesis.flyweight.Word;
import dk.rode.thesis.memento.MemorizableEnglishAlphabetSequence;
import dk.rode.thesis.memento.MemorizableSequence;
import dk.rode.thesis.memento.RangeSequence;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.DigitSequence;
import dk.rode.thesis.meta.model.HexDigitSequence;
import dk.rode.thesis.meta.model.LongSequence;
import dk.rode.thesis.meta.model.PrimeSequence;
import dk.rode.thesis.meta.model.ReversibleSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.observer.DateSequence;
import dk.rode.thesis.prototype.SymbolSequence;
import dk.rode.thesis.singleton.NorwegianAlphabetSequence;
import dk.rode.thesis.singleton.SmileySequence;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * <b>Visitor</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class Main implements Testable {

    /**
      * Executes the <b>Visitor</b> tests.
      * 
      * @param args The arguments to supply to {@link #test(Log, Arguments)},
      *             if any; can be null.
      */
    public static void main(String[] args) {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }
    
    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * 
      */
    public boolean test(Log out, Arguments arguments) {
        out.heading("Visitor tests...");
        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }

        // True for class logging...
        boolean log = arguments.getArgumentValue("log", true, Boolean.class);
        LogFactory.getLog(SimpleScanner.class).setActive(log);
        LogFactory.getLog(ReflectiveVisitableSequence.class).setActive(log);

        //
        // Create sequence structure
        //
        
        VisitableCompositeSequence cs1 = new VisitableCompositeSequence(CompositeStrategy.BREATH_FIRST);

        cs1.addSequence(new VisitableReversiblePrimeSequence(100));
        cs1.addSequence(new StringValuedVisitableSequence<CharSequence>(new DigitSequence()));
        cs1.addSequence(new DateValuedVisitableSequence(new DateSequence()));
        cs1.addSequence(new PrimeSequence(100));   
        cs1.addSequence(new VisitableLongSequence(25));
        cs1.addSequence(new VisitableRandomSequence(25));

        VisitableCompositeSequence cs2 = new VisitableCompositeSequence(CompositeStrategy.DEPTH_FIRST);
        cs2.addSequence(new StringValuedVisitableSequence<String>(new MemorizableEnglishAlphabetSequence()));
        cs2.addSequence(new MemorizableEnglishAlphabetSequence());
        
        cs1.addSequence(cs2);
        cs1.addSequence(new StringValuedVisitableSequence<CharSequence>(new SequenceAdapter<Long,CharSequence>(new LongSequence(), AdapterStrategy.NUMBER_TO_STRING)));
        cs1.addSequence(new RandomSequence(5));
        
        // cs3 and cs4 are used for traversal tests because of different 
        // strategies...
        CompositeSequence<CharSequence> cs3 = new CharSequenceCompositeSequence(CompositeStrategy.DEFAULT); // depth first
        cs3.addSequence(new SequenceAdapter<Date,CharSequence>(new DateSequence(new Date(), 5), AdapterStrategy.ANY_TO_STRING)); 
        cs3.addSequence(new StringValuedVisitableSequence<String>(NorwegianAlphabetSequence.getInstance()));

        VisitableCompositeSequence cs4 = new VisitableCompositeSequence(CompositeStrategy.BREATH_FIRST);
        cs4.addSequence(cs3);
        cs4.addSequence(new IntegerValuedVisitableSequence(new RangeSequence(100, 200))); 
        cs4.addSequence(new SymbolSequence());
        cs4.addSequence(new IntegerValuedVisitableSequence(new PrimeSequence(50))); 
        cs4.addSequence(new VisitableReversiblePrimeSequence(75));

        // Test reflective dispatch...
        cs4.addSequence(new ReflectiveVisitableSequence<Integer>(new RangeSequence(100, 200))); 
        cs4.addSequence(new ReflectiveVisitableSequence<CharSequence>(new HexDigitSequence()));
        
        // Deep structure...
        cs1.addSequence(cs4);
        
        CharacterFactory cf = new CharacterFactory();
        cs1.addSequence(new StringValuedVisitableSequence<Word>(cf.getSentence("Doh!")));
        cs1.addSequence(NorwegianAlphabetSequence.getInstance());
        cs1.addSequence(new StringValuedVisitableSequence<String>(SmileySequence.getSequence()));

        //
        // Show sequences used
        //
        
        out.heading("Created composites");

        Sequence<?>[] sequences = {cs1, cs2, cs3, cs4};
        
        for (int i = 0; i < sequences.length; i++) {
            Sequence<?> sequence = sequences[i];
            out.println("Composite sequence ", i, ": ");
            out.println(sequence);
            out.println("-> ", sequence.toString(CompositePolicy.VISIBLE));
            out.println();
        }
        
        //
        // Logging visitor (value and type)
        //
        
        out.heading("Using type and value visitor");

        LoggingVisitor lv = new LoggingVisitor(SequencePolicy.IDENTITY);

        out.println("Visitor: ", lv).println();
        
        CompositeStrategy strategy = CompositeStrategy.DEPTH_FIRST;
        SimpleScanner ss = new SimpleScanner(strategy);
                
        out.println("Type visitable scan 1/2...").println();
        ss.scan((TypeVisitableSequence<?>)cs1, lv, out);
        out.println();

        out.println("Value visitable scan 1/2...").println();
        ss.scan((ValueVisitableSequence<?>)cs1, lv, out);
        out.println();

        // Change strategies...
        ss.setCompositeStrategy(CompositeStrategy.BREATH_FIRST);
        lv.setSequencePolicy(SequencePolicy.ID);
        
        out.println("Type visitable scan 2/2...").println();
        ss.scan((TypeVisitableSequence<?>)cs1, lv, out);
        out.println();
        
        out.println("Value visitable scan 2/2...").println();
        ss.scan((ValueVisitableSequence<?>)cs1, lv, out);
        out.println();

        //
        // Counting visitor (value)
        //
        
        out.heading("Using value visitor");
        
        CountingVisitor cv = new CountingVisitor();
        
        out.println("Visitor: ", cv).println();
        
        out.println("Value visitable scan 1/1...").println();
        ss.scan(cs1, cv, null);
        
        out.println()
           .println("Scan result for *visited* sequences:")
           .println("------------------------------------")
           .println("Integer valued: ", cv.getIntegerValued())
           .println("Long valued:    ", cv.getLongValued())
           .println("Date valued:    ", cv.getDateValued())
           .println("String valued:  ", cv.getStringValued())
           .println("Composites:     ", cv.getComposites())
           .println("Unknowns:       ", cv.getUnknowns())
           .println()
           .println("Total:          ", cv.getTotal())
           .println();

        //
        // Type visitor (types)
        //
        
        ss.setCompositeStrategy(CompositeStrategy.DEPTH_FIRST);
        
        out.heading("Using type visitors");
        
        //
        // Composite
        //
        
        // Accept composite types only...
        TypeVisitor tv = new TypeVisitor() {
            @Override
            public void visitComposite(CompositeSequence<?> sequence, Void unused) {
                this.approve(sequence);
            }
        };
        out.println("Visitor (composite types only): ", tv).println();
        
        out.println("Type visitable scan 1/2...").println();
        ss.scan(cs1, tv, null);
        
        out.println("Sequences approved: ", tv.approved());
        for (Sequence<?> sequence : tv.getApproved()) {
            out.println("Sequence: ", sequence);
        }
        
        //
        // Memorizable or reversible
        //

        tv = new TypeVisitor() {
            @Override
            public void visitReversible(ReversibleSequence<?> sequence, Void unused) {
                this.approve(sequence);
            }
            @Override
            public void visitMemorizable(MemorizableSequence<?> sequence, Void unused) {
                this.approve(sequence);
            }
        };
        out.println();
        out.println("Visitor (reversible and memorizable types only): ", tv).println();
        
        out.println("Type visitable scan 2/2...").println();
        ss.scan(cs1, tv, null);
        
        out.println("Sequences approved: ", tv.approved());
        for (Sequence<?> sequence : tv.getApproved()) {
            out.println("Sequence: ", sequence);
        }

        out.heading("Visitor tests succeeded!");
        
        return true;
    }
       
}
