/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.visitor;

import dk.rode.thesis.composite.CompositeSequence;
import dk.rode.thesis.memento.MemorizableSequence;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.ReversibleSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.observer.ObservableSequence;
import dk.rode.thesis.state.StateableSequence;

/**
  * A <i>sequence type visitor</i> visits a {@link Sequence sequence}
  * based on its actual type, for example {@link ReversibleSequence}
  * or {@link MemorizableSequence}. <p>  
  *  
  * <b>Implementation notes</b>:<br/>
  * An actual visitor implementation of this interface may not be 
  * able to visit decorated, adapted, or proxied sequences, because
  * the {@code instanceof} test will fail in such cases. Such
  * sequences will thus use the {@link SequenceVisitor#visitUnknown(Sequence, Object)}
  * visitation method.
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    SequenceTypeScanner
  * @see    SequenceValueVisitor
  * 
  * @param <P> The type of the additional parameter to this visitor's methods.
  *            Use {@code Void} for visitors that do not need an additional 
  *            parameter.
  */
@Participant("Visitor")
public interface SequenceTypeVisitor<P> extends SequenceVisitor<P> {
    
    /**
      * Visits a {@link ReversibleSequence reversible} sequence.
      *  
      * @param  sequence The reversible sequence to visit; never null.
      * @param  argument A visitor-specified argument, if any; nullability
      *                  determined by the visitor implementation.
      */
    public void visitReversible(ReversibleSequence<?> sequence, P argument);

    /**
      * Visits a {@link MemorizableSequence memorizable} sequence.
      *  
      * @param  sequence The memorizable sequence to visit; never null.
      * @param  argument A visitor-specified argument, if any; nullability
      *                  determined by the visitor implementation.
      */
    public void visitMemorizable(MemorizableSequence<?> sequence, P argument);
    
    /**
      * Visits an {@link ObservableSequence observable} sequence.
      *  
      * @param  sequence The observable sequence to visit; never null.
      * @param  argument A visitor-specified argument, if any; nullability
      *                  determined by the visitor implementation.
      */
    public void visitObservable(ObservableSequence<?,?,?> sequence, P argument);

    /**
      * Visits a {@link StateableSequence stateable} sequence.
      *  
      * @param  sequence The stateable sequence to visit; never null.
      * @param  argument A visitor-specified argument, if any; nullability
      *                  determined by the visitor implementation.
      */
    public void visitStateable(StateableSequence<?> sequence, P argument);

    /**
      * Visits a {@link CompositeSequence composite} sequence.
      *  
      * @param  sequence The composite sequence to visit; never null.
      * @param  argument A visitor-specified argument, if any; nullability
      *                  determined by the visitor implementation.
      */
    public void visitComposite(CompositeSequence<?> sequence, P argument);
        
}
