/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.visitor;

import java.util.Date;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.Sequence;

/**
  * A <i>sequence value visitor</i> visits a {@link Sequence sequence}
  * based on the type of values it deliver, for example {@code Integer}
  * or {@code Date}. <p> 
  * 
  * For the type X, the <i>general</i> rule for the corresponding visitation 
  * method signature is {@code visitNameValued(Sequence<X>, Object)}, where 
  * {@code Name} is the simple name of {@code X}, for example 
  * {@link #visitLongValued(Sequence, Object) visitLongValued(Sequence<Long>, P)} 
  * for the {@code java.lang.Long} type. In case no such visitation method
  * can be found, the {@link SequenceVisitor#visitUnknown(Sequence, Object)}
  * method is used. <p>
  *
  * <b>Implementation notes</b>:<br/> 
  * Gamma et al. suggest the possibility of using {@code visit} as an 
  * overloaded visitation method name [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.337]. For this visitor 
  * this is <b>not</b> possible, since we differentiate on the <i>type 
  * parameter</i> used for the sequence argument, for example {@code Long} 
  * and {@code Date}, not the actual {@link Sequence} type itself. However, 
  * type parameters are erased at runtime, so it is only the <i>method 
  * name</i> that actually distinguishes the different visitation methods: 
  * at runtime, they all get a raw {@code Sequence} instance. Ordinary 
  * polymorphism can naturally still be applied, though, for overloaded 
  * method names. <p>
  * 
  * The visitation methods all accept an argument of type {@link Sequence}, 
  * not {@link ValueVisitableSequence}. This offers more flexibility and is 
  * reasonable since visitable sequences generally do not offer new sequence 
  * state, merely new behaviour. It also allows visitation to decorated, 
  * proxied, adapted, and composite sequences. Hence, the normal 
  * {@code Sequence} interface must be used to perform the actual
  * visitation logic as not to break encapsulation (see [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.337]).
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    SequenceValueScanner
  * @see    SequenceTypeVisitor
  * 
  * @param <P> The type of the additional parameter to this visitor's methods.
  *            Use {@code Void} for visitors that do not need an additional 
  *            parameter.
  */
@Participant("Visitor")
public interface SequenceValueVisitor<P> extends SequenceVisitor<P> {

    /**
      * Visits a {@link Sequence sequence} that delivers {@code Integer}
      * sequence values.
      *  
      * @param  sequence The sequence to visit; never null.
      * @param  argument A visitor-specified argument, if any; nullability
      *                  determined by the visitor implementation.
      */
    public void visitIntegerValued(Sequence<Integer> sequence, P argument);
    
    /**
      * Visits a {@link Sequence sequence} that delivers {@code Long}
      * sequence values.
      *  
      * @param  sequence The sequence to visit; never null.
      * @param  argument A visitor-specified argument, if any; nullability
      *                  determined by the visitor implementation.
      */
    public void visitLongValued(Sequence<Long> sequence, P argument);

    /**
      * Visits a {@link Sequence sequence} that delivers any type of 
      * value that implements the {@code CharSequence} interface,
      * including {@code CharSequence} it self. 
      *  
      * @param  sequence The sequence to visit; never null.
      * @param  argument A visitor-specified argument, if any; nullability
      *                  determined by the visitor implementation.
      */
    public void visitStringValued(Sequence<? extends CharSequence> sequence, P argument);
    
    /**
      * Visits a {@link Sequence sequence} that delivers {@code Date}
      * sequence values.
      *  
      * @param  sequence The sequence to visit; never null.
      * @param  argument A visitor-specified argument, if any; nullability
      *                  determined by the visitor implementation.
      */
    public void visitDateValued(Sequence<Date> sequence, P argument);
    
}
