/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.visitor;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import dk.rode.thesis.composite.CompositeSequence;
import dk.rode.thesis.memento.MemorizableSequence;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.ReversibleSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.observer.ObservableSequence;
import dk.rode.thesis.state.StateableSequence;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>type visitor</i> collects visited {@link Sequence sequences}
  * based on <i>acceptance</i> of specific sequence types. <p>
  * 
  * It handles only {@link TypeVisitableSequence type} visitable 
  * sequences. <p>
  * 
  * The collected types correspond to the visitation methods 
  * overridden by sub-classes of this visitor. If this visitor 
  * is used as is, it will accept no sequence types: each visitation 
  * method does nothing per default. A sub-class must simply invoke 
  * {@link #approve(Sequence)} to signal that a given visited
  * sequence is approved. <p> 
  * 
  * The visitor accumulates the visited sequences in visitation 
  * order. They can be collected after visitation via the 
  * {@link #getApproved()} method. A given sequence is only 
  * included once in the returned list, based on {@code equals},
  * not identity.   
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    LoggingVisitor
  * @see    CountingVisitor
  * @see    SimpleScanner
  */
@Participant("ConcreteVisitor")
public class TypeVisitor implements SequenceTypeVisitor<Void> {
    
    /**
      * The set containing the approved sequences, in order. <p>
      * 
      * Never null, but may be empty.
      */
    private final Set<Sequence<?>> approved; 
    
    /**
      * No-arg constructor. <p>
      */
    public TypeVisitor() {
        super();
        this.approved = new LinkedHashSet<Sequence<?>>();
    }
    
    /**
      * Clears the accumulated approved sequences, so this 
      * visitor can be reused. 
      */
    public void clear() {
        this.approved.clear();
    }
    
    /**
      * Returns a list containing the sequences approved by this
      * type visitor, in visitation order. <p>
      * 
      * Modifying the returned list will not affect this
      * visitor.
      *  
      * @return The list of sequences approved; never null, but can
      *         be empty.
      */
    public List<Sequence<?>> getApproved() {
        return new ArrayList<Sequence<?>>(this.approved);
    }

    /**
      * Returns the number of sequences approved by this visitor.
      *  
      * @return The number of sequences approved.
      * @see    #getApproved()
      */
    public int approved() {
        return this.approved.size();
    }
    
    /**
      * Approves the visited sequence supplied as {@code sequence}.
      * 
      * @param  sequence The sequence to approve; cannot be null.
      * @throws NullPointerException if {@code sequence} is null.
      */
    protected void approve(Sequence<?> sequence) {
        if (sequence == null) {
            throw new NullPointerException("Sequence cannot be null");
        }
        this.approved.add(sequence);
    }
    
    /**
      * 
      */
    public void visitReversible(ReversibleSequence<?> sequence, Void unused) {
        // Default implementation does nothing
    }
    
    /**
      * 
      */
    public void visitMemorizable(MemorizableSequence<?> sequence, Void unused) {
        // Default implementation does nothing
    }
    
    /**
      * 
      */
    public void visitObservable(ObservableSequence<?,?,?> sequence, Void unused) {
        // Default implementation does nothing
    }
    
    /**
      * 
      */
    public void visitStateable(StateableSequence<?> sequence, Void unused) {
        // Default implementation does nothing
    }
    
    /**
      * 
      */
    public void visitComposite(CompositeSequence<?> sequence, Void unused) {
        // Default implementation does nothing
    }

    /**
      * 
      */
    public void visitUnknown(Sequence<?> sequence, Void unused) {
        // Default implementation does nothing
    }
    
    /**
      * Returns the string representation of this visitor.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        // Cannot use simple class name, because that will generate weird
        // names when anonymous sub-classes are constructed, i.e. visitors
        // that only override for a single type or so...
        return new StringBuilder()
              .append("TypeVisitor[").append(ObjectPolicy.HASHCODE.toString(this))
              .append("; approved: ").append(this.approved.size())
              .append(']').toString();
    }
    
}
